# Copyright (c) 2012-2013 Turbulenz Limited

from jinja2 import Environment, FileSystemLoader, ChoiceLoader
from jinja2 import TemplateNotFound, TemplateSyntaxError, BaseLoader
from logging import getLogger

from turbulenz_tools.tools.toolsexception import ToolsException

import os
import re

LOG = getLogger(__name__)
LF_RE = re.compile(r".?\r.?", re.DOTALL)

############################################################

class DefaultTemplateLoader(BaseLoader):
    """
    Class that loads a named template from memory
    """

    def __init__(self, name, template):
        BaseLoader.__init__(self)
        self.name = name
        self.template = template

    def get_source(self, env, template):
        if template == self.name:
            LOG.info("Request for template '%s'.  Using default registered" \
                         % template)
            return self.template, None, lambda: True
        raise TemplateNotFound(template)

class UTF8FileSystemLoader(FileSystemLoader):
    """
    Class that loads a template from a given path, removing any utf8
    BOMs.
    """

    def __init__(self, searchpath):
        FileSystemLoader.__init__(self, searchpath)
        self._path = searchpath

    def get_source(self, env, name):
        fn = os.path.join(self._path, name)
        if not os.path.exists(fn):
            raise TemplateNotFound(name)

        d = read_file_utf8(fn)
        return d, None, lambda: True

############################################################

def _sanitize_crlf(m):
    v = m.group(0)
    if 3 == len(v):
        if '\n' == v[0]:
            return '\n' + v[2]
        if '\n' == v[2]:
            return v[0] + '\n'
    elif 2 == len(v):
        if '\n' in v:
            return '\n'
    return v.replace('\r', '\n')

# Read a file, handling any utf8 BOM
def read_file_utf8(filename):
    with open(filename, 'rb') as f:
        text = f.read().decode('utf-8-sig')

        # This is unpleasant, but it is thorough.  This is the
        # single-pass equivalent of:
        # return (in_data.replace('\r\n', '\n').replace('\n\r', '\n'))\
        #     .replace('\r', '\n');
        return LF_RE.sub(_sanitize_crlf, text)

############################################################

def env_create(options, default_template=None):
    """
    Setup a jinja env based on the tool options
    """

    LOG.info("Template dirs:")
    for t in options.templatedirs:
        LOG.info(" - '%s'" % t)

    loaders = [ UTF8FileSystemLoader(t) for t in options.templatedirs ]
    if default_template is not None:
        loaders.append(DefaultTemplateLoader('default', default_template))

    _loader = ChoiceLoader(loaders)
    env = Environment(loader = _loader,
                      block_start_string = '/*{%',
                      block_end_string = '%}*/',
                      variable_start_string = '/*{{',
                      variable_end_string = '}}*/',
                      comment_start_string = '/*{#',
                      comment_end_string = '#}*/')
    return env

############################################################

def env_load_template(env, in_file):
    """
    Load a single template into the environment, handling the
    appropriate errors.  Returns the loaded template
    """

    try:
        return env.get_template(in_file)
    except TemplateNotFound as e:
        raise ToolsException('template not found: %s' % str(e))
    except TemplateSyntaxError as e:
        raise ToolsException('template syntax error: %s' % str(e))

############################################################

def env_load_templates(env, inputs):
    """
    Load an array of templates into the environment.  Returns the
    loaded templates as a list
    """

    return [env_load_template(env, i) for i in inputs]



#!/usr/bin/python
# Copyright (c) 2010-2011,2013 Turbulenz Limited
"""
Convert Material Yaml (.material) files into a Turbulenz JSON asset.
"""

import logging
LOG = logging.getLogger('asset')

import yaml

# pylint: disable=W0403
from stdtool import standard_main, standard_json_out
from asset2json import JsonAsset
# pylint: enable=W0403

__version__ = '1.0.0'
__dependencies__ = ['asset2json']

#######################################################################################################################

def parse(input_filename="default.material", output_filename="default.json", asset_url="", asset_root=".",
          infiles=None, options=None):
    """
    Utility function to convert a Material Yaml (.material) into a JSON file.
    Known built-in textures are: default, quadratic, white, nofalloff, black, flat

    Example:

        # Example material
        material:
            effect: lambert
            diffuse: textures/wallstone.jpg
            color: [1.0, 0.5, 0.1]
            meta: &id1
                collision: True
                collisionFilter: ["ALL"]

        material2:
            diffuse: textures/wall.jpg
            meta:
                <<: *id1
                collision: False

        material3:
            diffuse: textures/stone.jpg
            meta: *id1
    """
    try:
        with open(input_filename, 'r') as source:
            try:
                materials = yaml.load(source)
            # pylint: disable=E1101
            except yaml.scanner.ScannerError as e:
            # pylint: enable=E1101
                LOG.error('Failed processing:%s' % input_filename)
                LOG.error('  >> %s' % e)
            else:
                json_asset = JsonAsset()
                for mat_name, material in materials.iteritems():
                    effect = material.pop('effect', None)
                    technique = material.pop('technique', None)
                    meta = material.pop('meta', None)
                    json_asset.attach_material(mat_name, effect, technique, material, meta)

                standard_json_out(json_asset, output_filename, options)
                return json_asset
    except IOError as e:
        LOG.error('Failed processing: %s' % output_filename)
        LOG.error('  >> %s' % e)
        return None

if __name__ == "__main__":
    try:
        standard_main(parse, __version__,
                      "Convert Material Yaml (.material) files into a Turbulenz JSON asset.",
                      __dependencies__)
    # pylint: disable=W0703
    except Exception as err:
        LOG.critical('Unexpected exception: %s' % err)
        exit(1)
    # pylint: enable=W0703



# Copyright (c) 2009-2013 Turbulenz Limited
"""
Asset class used to build Turbulenz JSON assets.
"""

import logging
LOG = logging.getLogger('asset')

from itertools import chain as itertools_chain
from simplejson import encoder as json_encoder, dumps as json_dumps, dump as json_dump
from types import StringType

# pylint: disable=W0403
import vmath

from turbulenz_tools.utils.json_utils import float_to_string, metrics
from node import NodeName
from material import Material
# pylint: enable=W0403

__version__ = '1.2.0'
__dependencies__ = ['vmath', 'json2json', 'node', 'material']

#######################################################################################################################

def attach_skins_and_materials(asset, definitions, material_name, default=True):
    """Find all skins in the ``definitions`` asset which remap the material ``material_name``. Attach the found skin
    and references materials to the target ``asset``."""
    skins = definitions.retrieve_skins()
    for skin, materials in skins.iteritems():
        attach_skin = False
        for v, k in materials.iteritems():
            if v == material_name:
                attach_skin = True
                material = definitions.retrieve_material(k, default)
                asset.attach_material(k, raw=material)
        if attach_skin:
            asset.attach_skin(skin, materials)

def remove_unreferenced_images(asset, default=True):
    def _unreference_image(imageName):
        if type(imageName) is StringType:
            # this image is referenced so remove it from the unreferenced list
            if imageName in unreferenced_images:
                unreferenced_images.remove(imageName)

    images = asset.asset['images']
    unreferenced_images = set(asset.asset['images'].iterkeys())
    materials = asset.asset['materials']
    effects = asset.asset['effects']

    for v, _ in effects.iteritems():
        effect = asset.retrieve_effect(v)
        parameters = effect.get('parameters', None)
        if parameters:
            for value in parameters.itervalues():
                _unreference_image(value)

    for v, _ in materials.iteritems():
        material = asset.retrieve_material(v, default)
        parameters = material.get('parameters', None)
        if parameters:
            for value in parameters.itervalues():
                _unreference_image(value)

    # remove the unreferenced images
    for i in unreferenced_images:
        del images[i]

#######################################################################################################################

DEFAULT_IMAGE_FILENAME = 'default.png'
DEFAULT_EFFECT_TYPE = 'lambert'
DEFAULT_EFFECT_NAME = 'effect-0'
DEFAULT_MATERIAL_NAME = 'material-0'
DEFAULT_SHAPE_NAME = 'shape-0'
DEFAULT_SKELETON_NAME = 'skeleton-0'
DEFAULT_LIGHT_NAME = 'light-0'
DEFAULT_INSTANCE_NAME = 'instance-0'
DEFAULT_NODE_NAME = NodeName('node-0')
DEFAULT_TEXTURE_MAP_NAME = 'shape-map1'
DEFAULT_ANIMATION_NAME = 'animation-0'
DEFAULT_ENTITY_DEFINITION_NAME = 'entity_definition-0'
DEFAULT_MODEL_DEFINITION_NAME = 'model_definition-0'
DEFAULT_ENTITY_NAME = 'entity-0'
DEFAULT_PHYSICS_MATERIAL_NAME = 'physics-material-0'
DEFAULT_PHYSICS_NODE_NAME = 'physics-node-0'
DEFAULT_PHYSICS_MODEL_NAME = 'physics-model-0'
DEFAULT_SOUND_NAME = 'sound-0'
DEFAULT_SKIN_NAME = 'skin-0'
DEFAULT_PROCEDURALEFFECT_NAME = 'effect-0'
DEFAULT_APPLICATION_NAME = 'player'

# pylint: disable=R0904
class JsonAsset():
    """Contains a JSON asset."""
    SurfaceLines = 0
    SurfaceTriangles = 1
    SurfaceQuads = 2

    def __init__(self, up_axis='Y', v = 1, definitions=None):
        if definitions:
            self.asset = definitions
        else:
            self.asset = { 'version': v,
                           'geometries': { },
                           'skeletons': { },
                           'effects': { },
                           'materials': { },
                           'images': { },
                           'nodes': { },
                           'lights': { },
                           'entitydefinitions': { },
                           'modeldefinitions': { },
                           'entities': { },
                           'animations': { },
                           'camera_animations': { },
                           'physicsmaterials': { },
                           'physicsmodels': { },
                           'physicsnodes': { },
                           'sounds': { },
                           'proceduraleffects': { },
                           'areas': [ ],
                           'bspnodes': [ ],
                           'skins': { },
                           'strings': { },
                           'guis': { },
                           'tables': { },
                           'applications': { }
                         }
        if up_axis == 'X':
            self.default_transform = [ 0, 1, 0, -1, 0,  0, 0, 0, 1, 0, 0, 0 ]
        elif up_axis == 'Y':
            self.default_transform = [ 1, 0, 0,  0, 1,  0, 0, 0, 1, 0, 0, 0 ]
        elif up_axis == 'Z':
            self.default_transform = [ 1, 0, 0,  0, 0, -1, 0, 1, 0, 0, 0, 0 ]

    def json_to_string(self, sort=True, indent=1):
        """Convert the asset to JSON and return it as a string."""
        json_encoder.FLOAT_REPR = float_to_string
        return json_dumps(self.asset, sort_keys=sort, indent=indent)

    def json_to_file(self, target, sort=True, indent=0):
        """Convert the asset to JSON and write it to the file stream."""
        json_encoder.FLOAT_REPR = float_to_string
        if indent > 0:
            return json_dump(self.asset, target, sort_keys=sort, indent=indent)
        else:
            return json_dump(self.asset, target, sort_keys=sort, separators=(',', ':'))

    def clean(self):
        """Remove any toplevel elements which are empty."""
        for k in self.asset.keys():
            if isinstance(self.asset[k], dict) and len(self.asset[k].keys()) == 0:
                del self.asset[k]
            elif isinstance(self.asset[k], list) and len(self.asset[k]) == 0:
                del self.asset[k]

    def log_metrics(self):
        """Output the metrics to the log."""
        m = metrics(self.asset)
        keys = m.keys()
        keys.sort()
        for k in keys:
            LOG.info('json_asset:%s:%s' % (k, m[k]))

#######################################################################################################################

    def __set_source(self, shape, name, stride, min_element=None, max_element=None, data=None):
        """Add a vertex stream source for the specified geometry shape."""
        if data is None:
            data = [ ]
        source = { 'stride': stride, 'data': data }
        if min_element is not None:
            source['min'] = min_element
        if max_element is not None:
            source['max'] = max_element
        LOG.debug("geometries:%s:sources:%s[%i]" % (shape, name, len(data) / stride))
        self.asset['geometries'][shape]['sources'][name] = source

    def __set_input(self, shape, element, data):
        """Add a vertex stream input for the specified geometry shape."""
        LOG.debug("geometries:%s:inputs:%s:%s offset %i" % (shape, element, data['source'], data['offset']))
        self.asset['geometries'][shape]['inputs'][element] = data

    def __set_shape(self, shape, key, data, name=None):
        """Add a key and data to the specified geometry shape."""
        if isinstance(data, list):
            LOG.debug("geometries:%s:%s[%i]" % (shape, key, len(data)))
        else:
            LOG.debug("geometries:%s:%s:%i" % (shape, key, data))
        if name is None:
            self.asset['geometries'][shape][key] = data
        else:
            shape_asset = self.asset['geometries'][shape]
            if 'surfaces' not in shape_asset:
                shape_asset['surfaces'] = { }
            surfaces_asset = shape_asset['surfaces']
            if name not in surfaces_asset:
                surfaces_asset[name] = { }
            surface_asset = surfaces_asset[name]
            surface_asset[key] = data

    def __set_meta(self, shape, data):
        """Add the meta information to the specified geometry shape."""
        self.asset['geometries'][shape]['meta'] = data

    def __set_geometry(self, key, data):
        """Add a key and data to single geometry."""
        LOG.debug("geometries:%s:%s" % (key, data))
        self.asset['geometries'][key] = data

    def __set_asset(self, key, data):
        """Add a key and data to the asset."""
        LOG.debug("%s:%s" % (key, data))
        self.asset[key] = data

    def __attach_v1(self, shape, source, name, attribute, offset=0):
        """Attach a single value stream to the JSON representation. Also calculates the min and max range."""
        elements = [ ]
        (min_x) = source[0]
        (max_x) = source[0]
        for x in source:
            elements.append(x)
            min_x = min(x, min_x)
            max_x = max(x, max_x)
        self.__set_source(shape, name, 1, [min_x], [max_x], elements)
        self.__set_input(shape, attribute, { 'source': name, 'offset': offset })

    def __attach_v2(self, shape, source, name, attribute, offset=0):
        """Attach a Tuple[2] stream to the JSON representation. Also calculates the min and max range."""
        elements = [ ]
        (min_x, min_y) = source[0]
        (max_x, max_y) = source[0]
        for (x, y) in source:
            elements.append(x)
            elements.append(y)
            min_x = min(x, min_x)
            min_y = min(y, min_y)
            max_x = max(x, max_x)
            max_y = max(y, max_y)
        self.__set_source(shape, name, 2, [min_x, min_y], [max_x, max_y], elements)
        self.__set_input(shape, attribute, { 'source': name, 'offset': offset })

    def __attach_v3(self, shape, source, name, attribute, offset=0):
        """Attach a Tuple[3] stream to the JSON representation. Also calculates the min and max range."""
        if len(source) > 0:
            elements = [ ]
            (min_x, min_y, min_z) = source[0]
            (max_x, max_y, max_z) = source[0]
            for (x, y, z) in source:
                elements.append(x)
                elements.append(y)
                elements.append(z)
                min_x = min(x, min_x)
                min_y = min(y, min_y)
                min_z = min(z, min_z)
                max_x = max(x, max_x)
                max_y = max(y, max_y)
                max_z = max(z, max_z)
            self.__set_source(shape, name, 3, [min_x, min_y, min_z], [max_x, max_y, max_z], elements)
            self.__set_input(shape, attribute, { 'source': name, 'offset': offset })

    def __attach_v4(self, shape, source, name, attribute, offset=0):
        """Attach a Tuple[4] stream to the JSON representation. Also calculates the min and max range."""
        elements = [ ]
        (min_x, min_y, min_z, min_w) = source[0]
        (max_x, max_y, max_z, max_w) = source[0]
        for (x, y, z, w) in source:
            elements.append(x)
            elements.append(y)
            elements.append(z)
            elements.append(w)
            min_x = min(x, min_x)
            min_y = min(y, min_y)
            min_z = min(z, min_z)
            min_w = min(w, min_w)
            max_x = max(x, max_x)
            max_y = max(y, max_y)
            max_z = max(z, max_z)
            max_w = max(w, max_w)
        self.__set_source(shape, name, 4, [min_x, min_y, min_z, min_w], [max_x, max_y, max_z, max_w], elements)
        self.__set_input(shape, attribute, { 'source': name, 'offset': offset })

    def __retrieve_node(self, name):
        """Find the named node in the node hierarchy."""
        parts = name.hierarchy_names()
        node = self.asset
        for p in parts:
            if 'nodes' not in node:
                node['nodes'] = { }
            nodes = node['nodes']
            if p not in nodes:
                nodes[p] = { }
            node = nodes[p]
        return node

    def __retrieve_shape_instance(self,
                                  name=DEFAULT_NODE_NAME,
                                  shape_instance=DEFAULT_INSTANCE_NAME):
        assert(isinstance(name, NodeName))
        node = self.__retrieve_node(name)
        if 'geometryinstances' in node:
            geometry_instances = node['geometryinstances']
            if shape_instance in geometry_instances:
                instance = geometry_instances[shape_instance]
            else:
                instance = { }
                geometry_instances[shape_instance] = instance
        else:
            geometry_instances = { }
            node['geometryinstances'] = geometry_instances
            instance = { }
            geometry_instances[shape_instance] = instance
        return instance

    def __retrieve_light_instance(self,
                                  name=DEFAULT_NODE_NAME,
                                  light_instance=DEFAULT_INSTANCE_NAME):
        assert(isinstance(name, NodeName))
        node = self.__retrieve_node(name)
        if 'lightinstances' in node:
            light_instances = node['lightinstances']
            if light_instance in light_instances:
                instance = light_instances[light_instance]
            else:
                instance = { }
                light_instances[light_instance] = instance
        else:
            light_instances = { }
            node['lightinstances'] = light_instances
            instance = { }
            light_instances[light_instance] = instance
        return instance

#######################################################################################################################

    def attach_stream(self, source, shape, name, semantic, stride, offset):
        """Missing"""
        if source is not None:
            if isinstance(source[0], tuple) is False:
                self.__attach_v1(shape, source, name, semantic, offset)
            elif len(source[0]) == 2:
                self.__attach_v2(shape, source, name, semantic, offset)
            elif len(source[0]) == 3:
                self.__attach_v3(shape, source, name, semantic, offset)
            elif len(source[0]) == 4:
                self.__attach_v4(shape, source, name, semantic, offset)
        else:
            self.__set_source(shape, name, stride)
            self.__set_input(shape, semantic, { 'source': name, 'offset': offset })

    def attach_uvs(self, uvs, shape=DEFAULT_SHAPE_NAME, name=DEFAULT_TEXTURE_MAP_NAME, semantic="TEXCOORD0"):
        """Attach the vertex UVs to the JSON representation. Also calculates the min and max range."""
        if len(uvs[0]) == 2:
            self.__attach_v2(shape, uvs, name, semantic)
        elif len(uvs[0]) == 3:
            self.__attach_v3(shape, uvs, name, semantic)

    def attach_positions(self, positions, shape=DEFAULT_SHAPE_NAME, name="shape-positions", semantic="POSITION"):
        """Attach the vertex position stream to the JSON representation. Also calculates the min and max range."""
        self.__attach_v3(shape, positions, name, semantic)

    def attach_normals(self, normals, shape=DEFAULT_SHAPE_NAME, name="shape-normals", semantic="NORMAL"):
        """Attach the vertex normal stream to the JSON representation. Also calculates the min and max range."""
        self.__attach_v3(shape, normals, name, semantic)

    # pylint: disable=R0913
    def attach_nbts(self, normals, tangents, binormals, shape=DEFAULT_SHAPE_NAME,
                    normals_name="shape-normals", tangents_name="shape-tangents", binormals_name="shape-binormals",
                    normals_semantic="NORMAL", tangents_semantic="TANGENT", binormals_semantic="BINORMAL"):
        """Attach the vertex nbt streama to the JSON representation. Also calculates the min and max range."""
        self.__attach_v3(shape, normals, normals_name, normals_semantic)
        self.__attach_v3(shape, tangents, tangents_name, tangents_semantic)
        self.__attach_v3(shape, binormals, binormals_name, binormals_semantic)
    # pylint: enable=R0913

    def attach_skinning_data(self, skin_indices, skin_weights, shape=DEFAULT_SHAPE_NAME,
                             skin_indices_name="shape-skinindices", skin_indices_semantic="BLENDINDICES",
                             skin_weights_name="shape-skinweights", skin_weights_semantic="BLENDWEIGHT"):
        """Attach the vertex skinning indices and weights streams to the JSON representation.
        Also calculates the min and max range."""
        self.__attach_v4(shape,  skin_indices, skin_indices_name, skin_indices_semantic)
        self.__attach_v4(shape,  skin_weights, skin_weights_name, skin_weights_semantic)

    def attach_shape(self, shape=DEFAULT_SHAPE_NAME):
        """Attach the shapes to the JSON representation. This should be done before adding any shape streams."""
        self.__set_geometry(shape, { 'sources': { }, 'inputs':  { } })

    def attach_meta(self, meta, shape=DEFAULT_SHAPE_NAME):
        """Attach the meta information to the JSON representation of the shape."""
        self.__set_meta(shape, meta)

    def attach_surface(self, primitives, primitive_type, shape=DEFAULT_SHAPE_NAME, name=None):
        """Attach a surface to the JSON representation. Primitive type should be:
                SurfaceLines = 0
                SurfaceTriangles = 1
                SurfaceQuads = 2
        The primitives will be added to the specified `shape`.

        If a `name` is also specified then the primitives will be put into a named surfaces dictionary."""
        # Collapse the primitives down into a flat index list
        num_primitives = len(primitives)
        indices = [ ]
        for p in primitives:
            indices.extend(p)
        if 0 == len(indices):
            LOG.error('No indices for %s on %s', name, shape)
        if isinstance(indices[0], (tuple, list)):
            indices = list(itertools_chain(*indices))

        self.__set_shape(shape, 'numPrimitives', num_primitives, name)
        if primitive_type == JsonAsset.SurfaceLines:
            self.__set_shape(shape, 'lines', indices, name)
        elif primitive_type == JsonAsset.SurfaceTriangles:
            self.__set_shape(shape, 'triangles', indices, name)
        elif primitive_type == JsonAsset.SurfaceQuads:
            self.__set_shape(shape, 'quads', indices, name)
        else:
            LOG.error('Unsupported primitive type:%i' % primitive_type)

    def attach_geometry_skeleton(self, shape=DEFAULT_SHAPE_NAME, skeleton=DEFAULT_SKELETON_NAME):
        """Add a skeleton for the specified geometry shape."""
        LOG.debug("geometries:%s:skeleton added:%s" % (shape, skeleton))
        self.asset['geometries'][shape]['skeleton'] = skeleton

    def attach_skeleton(self, skeleton, name=DEFAULT_SKELETON_NAME):
        """Add a skeleton object."""
        LOG.debug("%s:skeleton added" % (name))
        self.asset['skeletons'][name] = skeleton

    def attach_bbox(self, bbox):
        """Attach the bounding box to the top-level geometry of the JSON representation."""
        self.__set_asset('min', bbox['min'])
        self.__set_asset('max', bbox['max'])

    def attach_image(self, filename=DEFAULT_IMAGE_FILENAME, image_link=None):
        """Attach an image to the JSON respresentation."""
        if image_link is None:
            for image_link, image_filename in self.asset['images'].iteritems():
                if image_filename == filename:
                    return image_link
            index = len(self.asset['images'])
            image_link = "file%03i" % index
        self.asset['images'][image_link] = filename
        return image_link

    def attach_effect(self, name=DEFAULT_EFFECT_NAME, effect_type=DEFAULT_EFFECT_TYPE,
                      parameters=None, shader=None, meta=None, raw=None):
        """Attach a new effect to the JSON representation."""
        if raw:
            LOG.debug("effects:%s:elements:%i" % (name, len(raw.keys())))
            self.asset['effects'][name] = raw
        else:
            if parameters is None:
                parameters = { }
            if name not in self.asset['effects']:
                LOG.debug("effects:%s:type:%s" % (name, effect_type))
                effect = { 'type': effect_type, 'parameters': parameters }
                if shader is not None:
                    effect['shader'] = shader
                if meta is not None:
                    effect['meta'] = meta
                self.asset['effects'][name] = effect

    def attach_material(self, name=DEFAULT_MATERIAL_NAME, effect=DEFAULT_EFFECT_NAME,
                        technique=None, parameters=None, meta=None, raw=None):
        """Attach a material to the JSON representation."""
        if raw:
            if 'stages' in raw:
                num_stages = len(raw['stages'])
            else:
                num_stages = 0
            LOG.debug("materials:%s:elements:%i:stages:%i" % (name, len(raw.keys()), num_stages))
            self.asset['materials'][name] = raw
        else:
            LOG.debug("materials:%s:effect:%s" % (name, effect))
            material = { }
            if effect is not None:
                material['effect'] = effect
            if technique is not None:
                material['technique'] = technique
            if parameters is not None:
                material['parameters'] = parameters
            if meta is not None:
                material['meta'] = meta
            self.asset['materials'][name] = material

    def retrieve_effect(self, name):
        """Return a reference to an effect."""
        if 'effects' in self.asset and name in self.asset['effects']:
            return Material(self.asset['effects'][name])
        return None

    def retrieve_material(self, name=DEFAULT_MATERIAL_NAME, default=True):
        """Return a reference to a material."""
        if name in self.asset['materials']:
            return Material(self.asset['materials'][name])

        if default:
            default_material = { 'effect': 'default',
                                 'parameters': { 'diffuse': 'default' },
                                 'meta': { 'tangents': True } }
            LOG.warning("Material not found:%s", name)
            return Material(default_material)

        return None

    def attach_texture(self, material, stage, filename):
        """Attach an image and linked parameter to the effect of the JSON representation."""
        assert( material in self.asset['materials'] )
        file_link = self.attach_image(filename)
        LOG.debug("material:%s:parameters:%s:%s" % (material, stage, filename))
        # Override the texture definition to redirect to this new shortcut to the image
        self.asset['materials'][material]['parameters'][stage] = file_link

    def attach_node(self, name=DEFAULT_NODE_NAME, transform=None):
        """Attach a node with a transform to the JSON representation."""
        assert(isinstance(name, NodeName))
        if not transform:
            transform = self.default_transform
        node = self.__retrieve_node(name)
        LOG.debug("nodes:%s:matrix:%s" % (name, transform))
        if len(transform) == 16:
            transform = vmath.m43from_m44(transform)
        if not vmath.m43is_identity(transform):
            node['matrix'] = transform

    def attach_node_shape_instance(self, name=DEFAULT_NODE_NAME,
                                   shape_instance=DEFAULT_INSTANCE_NAME,
                                   shape=DEFAULT_SHAPE_NAME,
                                   material=DEFAULT_MATERIAL_NAME,
                                   surface=None,
                                   disabled=False):
        """Attach a node connecting a shape, material and transform to the JSON representation."""
        assert(isinstance(name, NodeName))
        assert(shape in self.asset['geometries'])
        LOG.debug("nodes:%s:geometry:%s" % (name, shape))
        if not material in self.asset['materials']:
            LOG.info("nodes:%s:referencing missing material:%s" % (name, material))
        else:
            LOG.debug("nodes:%s:material:%s" % (name, material))
        LOG.debug("nodes:%s:disabled:%s" % (name, disabled))
        instance = self.__retrieve_shape_instance(name, shape_instance)
        instance['geometry'] = shape
        instance['material'] = material
        if surface is not None:
            instance['surface'] = surface
        if disabled:
            instance['disabled'] = disabled

    def attach_shape_instance_attributes(self,
                                         name=DEFAULT_NODE_NAME,
                                         shape_instance=DEFAULT_INSTANCE_NAME,
                                         attributes=None):
        """Copy the attributes onto the node."""
        assert(isinstance(name, NodeName))
        instance = self.__retrieve_shape_instance(name, shape_instance)

        if attributes:
            for k, v in attributes.iteritems():
                instance[k] = v

    def attach_shape_instance_material(self,
                                       name=DEFAULT_NODE_NAME,
                                       shape_instance=DEFAULT_INSTANCE_NAME,
                                       material=DEFAULT_MATERIAL_NAME):
        """Attach a node connecting a material to the JSON representation."""
        assert(isinstance(name, NodeName))
        if not material in self.asset['materials']:
            LOG.info("nodes:%s:referencing missing material:%s" % (name, material))
        else:
            LOG.debug("nodes:%s:material:%s" % (name, material))
        instance = self.__retrieve_shape_instance(name, shape_instance)
        instance['material'] = material

    def attach_node_light_instance(self, name=DEFAULT_NODE_NAME,
                                   light_instance=DEFAULT_INSTANCE_NAME,
                                   light=DEFAULT_LIGHT_NAME,
                                   disabled=False):
        """Attach a node connecting a light and transform to the JSON representation."""
        assert(isinstance(name, NodeName))
        assert(light in self.asset['lights'])
        LOG.debug("nodes:%s:light:%s" % (name, light))
        instance = self.__retrieve_light_instance(name, light_instance)
        instance['light'] = light
        if disabled:
            instance['disabled'] = disabled

    def attach_node_attributes(self, name=DEFAULT_NODE_NAME, attributes=None):
        """Copy the attributes onto the node."""
        assert(isinstance(name, NodeName))
        node = self.__retrieve_node(name)
        if attributes:
            for k, v in attributes.iteritems():
                node[k] = v

#######################################################################################################################

    def attach_area(self, name):
        """Attach an area targetting a node."""
        area = { 'target': name, 'portals': [ ] }
        if not name in self.asset['nodes']:
            LOG.warning("portal:%s:referencing missing node:%s", name, name)
        self.asset['areas'].append(area)

    def attach_area_portal(self, index, target_index, points):
        """Attach an area portal to the area. This contains a target area, a target node and the portal points."""
        ((min_x, min_y, min_z), (max_x, max_y, max_z)) = vmath.v3s_min_max(points)
        center = ( (max_x + min_x) / 2, (max_y + min_y) / 2, (max_z + min_z) / 2 )
        halfextents = ( (max_x - min_x) / 2, (max_y - min_y) / 2, (max_z - min_z) / 2 )
        portal = { 'area': target_index, 'points': points, 'center': center, 'halfExtents': halfextents }
        self.asset['areas'][index]['portals'].append(portal)

    def attach_bsp_tree_node(self, plane, pos, neg):
        """Attach a bsp tree node."""
        node = { 'plane':plane, 'pos':pos, 'neg':neg }
        self.asset['bspnodes'].append(node)

    def retrieve_light(self, name):
        """Return a reference to a light."""
        if 'lights' in self.asset and name in self.asset['lights']:
            return self.asset['lights'][name]
        return None

    def attach_light(self, name=DEFAULT_LIGHT_NAME, raw=None):
        """Attach a light to the JSON representation."""
        self.asset['lights'][name] = raw

#######################################################################################################################

    def attach_entity_definition(self, name=DEFAULT_ENTITY_DEFINITION_NAME, attributes=None):
        """Attach an entity definition to the JSON representation."""
        if not attributes:
            attributes = { }
        self.asset['entitydefinitions'][name] = attributes

    def attach_model_definition(self, name=DEFAULT_MODEL_DEFINITION_NAME, attributes=None):
        """Attach an model definition to the JSON representation."""
        if not attributes:
            attributes = { }
        self.asset['modeldefinitions'][name] = attributes

    def attach_entity(self, name=DEFAULT_ENTITY_NAME, attributes=None):
        """Attach an entity to the JSON representation."""
        if not attributes:
            attributes = { }
        self.asset['entities'][name] = attributes

    def attach_skin(self, name=DEFAULT_SKIN_NAME, attributes=None):
        """Attach a skin to the JSON representation."""
        if not attributes:
            attributes = { }
        self.asset['skins'][name] = attributes

    def retrieve_skins(self):
        """Return all the skins."""
        return self.asset.get('skins', { })

    def retrieve_skin(self, name=DEFAULT_SKIN_NAME):
        """Return a reference to a skin."""
        skins = self.retrieve_skins()
        if name in skins:
            skin = skins[name]
        else:
            skin = { }
        return skin

#######################################################################################################################

    def attach_animation(self, name=DEFAULT_ANIMATION_NAME, animation=None):
        """Attach an animation to the JSON representation."""
        self.asset['animations'][name] = animation

    def attach_camera_animation(self, name=DEFAULT_ANIMATION_NAME, animation=None):
        """Attach a camera animation to the JSON representation."""
        self.asset['camera_animations'][name] = animation

#######################################################################################################################

    def attach_sound(self, name=DEFAULT_SOUND_NAME, sound=None):
        """Attach a sound to the JSON representation."""
        self.asset['sounds'][name] = sound

    def attach_proceduraleffect(self, name=DEFAULT_PROCEDURALEFFECT_NAME, proceduraleffect=None):
        """Attach a proceduraleffect to the JSON representation."""
        self.asset['proceduraleffects'][name] = proceduraleffect

#######################################################################################################################

    def attach_physics_material(self, physics_material_name=DEFAULT_PHYSICS_MATERIAL_NAME,
                                      physics_material=None):
        self.asset['physicsmaterials'][physics_material_name] = physics_material

    def attach_physics_model(self, physics_model=DEFAULT_PHYSICS_MODEL_NAME,
                                   shape_name=DEFAULT_SHAPE_NAME, model=None, material=None,
                                   shape_type="mesh"):
        """Attach a physics model to the JSON representation."""
        if model is None:
            model = { 'type': "rigid",
                      'shape': shape_type,
                      'geometry': shape_name }
        if material:
            model['material'] = material
        self.asset['physicsmodels'][physics_model] = model

    def attach_physics_node(self, physics_node_name=DEFAULT_PHYSICS_NODE_NAME,
                                  physics_model=DEFAULT_PHYSICS_MODEL_NAME,
                                  node_name=DEFAULT_NODE_NAME,
                                  inline_parameters=None):
        """Attach a physics node to the JSON representation."""
        assert(isinstance(node_name, NodeName))
        physics_node = { 'body': physics_model, 'target': str(node_name) }
        if inline_parameters is not None:
            for k, v in inline_parameters.iteritems():
                physics_node[k] = v
        self.asset['physicsnodes'][physics_node_name] = physics_node

#######################################################################################################################

    def attach_strings(self, name, strings):
        """Attach a named string table."""
        self.asset['strings'][name] = strings

    def attach_guis(self, name, gui):
        """Attach a named window."""
        if name not in self.asset['guis']:
            self.asset['guis'][name] = [ ]
        self.asset['guis'][name].append(gui)

    def attach_table(self, name, table):
        """Attach a table."""
        self.asset['tables'][name] = table

    def retrieve_table(self, name):
        """Retrieve a reference to a table."""
        return self.asset['tables'][name]

    def attach_application(self, name=DEFAULT_APPLICATION_NAME, options=None):
        """Attach options for an application."""
        if options is None:
            options = { }
        self.asset['applications'][name] = options
# pylint: enable=R0904



# This code is based on an xml2json.py script found at:
# https://gist.github.com/raw/434945/2a0615b2bd07ece2248a968609284b3ba0d5e466/xml2json.py
#
# It has been modified to fit our package style and asset types.
# Also supports stripping out namespaces and converting values to native types.
#
# All modifications are:
# Copyright (c) 2010-2011,2013 Turbulenz Limited

from simplejson import loads as json_loads, dumps as json_dumps, encoder as json_encoder

# pylint: disable=W0404
try:
    from xml.etree import cElementTree as ElementTree
except ImportError:
    from xml.etree import ElementTree
# pylint: enable=W0404

from turbulenz_tools.tools.json2json import float_to_string

__version__ = '1.0.0'
__dependencies__ = ['turbulenz_tools.tools.json2json']

#######################################################################################################################

def to_native(a):
    """Parse the string into native types."""
    if a is None:
        return None

    try:
        int_number = int(a)
    except ValueError:
        pass
    else:
        return int_number

    try:
        float_number = float(a)
    except ValueError:
        pass
    else:
        return float_number

    parts = a.split()

    try:
        int_list = [int(x) for x in parts]
    except ValueError:
        pass
    else:
        return int_list

    try:
        float_list = [float(x) for x in parts]
    except ValueError:
        pass
    else:
        return float_list

    return a

def elem2internal(elem, strip=True, convert_types=False):
    """Convert an Element into an internal dictionary (not JSON!)."""
    if convert_types:
        _prepare = to_native
    else:
        _prepare = lambda a: a

    def _elem2internal(elem):
        d = { }
        for key, value in elem.attrib.items():
            d['@'+key] = _prepare(value)

        # loop over subelements to merge them
        for subelem in elem:
            v = _elem2internal(subelem)
            tag = subelem.tag
            value = v[tag]
            try:
                # add to existing list for this tag
                d[tag].append(value)
            except AttributeError:
                # turn existing entry into a list
                d[tag] = [d[tag], value]
            except KeyError:
                # add a new non-list entry
                d[tag] = value
        text = elem.text
        tail = elem.tail
        if strip:
            # ignore leading and trailing whitespace
            if text:
                text = text.strip()
            if tail:
                tail = tail.strip()
        text = _prepare(text)
        tail = _prepare(tail)

        if tail:
            d['#tail'] = tail

        if d:
            # use #text element if other attributes exist
            if text:
                d["#text"] = text
        else:
            # text is the value if no attributes

            # The following line used to read:
            #  >> d = text or None
            # But we now convert '0' to 0 and this resulted in None instead of 0.
            # So it has been updated to:
            d = text
        return { elem.tag: d }

    return _elem2internal(elem)

def internal2elem(pfsh, factory=ElementTree.Element):
    """Convert an internal dictionary (not JSON!) into an Element."""
    attribs = { }
    text = None
    tail = None
    sublist = [ ]
    tag = pfsh.keys()
    if len(tag) != 1:
        raise ValueError("Illegal structure with multiple tags: %s" % tag)
    tag = tag[0]
    value = pfsh[tag]
    if isinstance(value, dict):
        for k, v in value.items():
            if k[:1] == "@":
                attribs[k[1:]] = v
            elif k == "#text":
                text = v
            elif k == "#tail":
                tail = v
            elif isinstance(v, list):
                for v2 in v:
                    sublist.append(internal2elem({k:v2}, factory=factory))
            else:
                sublist.append(internal2elem({k:v}, factory=factory))
    else:
        text = value
    e = factory(tag, attribs)
    for sub in sublist:
        e.append(sub)
    e.text = text
    e.tail = tail
    return e

def elem2json(elem, strip=True, indent=0, convert_types=False):
    """Convert an ElementTree or Element into a JSON string."""
    if hasattr(elem, 'getroot'):
        elem = elem.getroot()

    internal = elem2internal(elem, strip=strip, convert_types=convert_types)

    # Module 'simplejson' has no 'encoder' member
    # pylint: disable=E1101
    json_encoder.FLOAT_REPR = float_to_string
    # pylint: enable=E1101
    if indent > 0:
        output = json_dumps(internal, sort_keys=True, indent=indent)
    else:
        output = json_dumps(internal, sort_keys=True, separators=(',', ':'))

    return output

def json2elem(json_string, factory=ElementTree.Element):
    """Convert a JSON string into an Element."""
    return internal2elem(json_loads(json_string), factory)

def xml2json(xml_string, strip=True, indent=0, convert_types=False):
    """Convert an XML string into a JSON string."""
    elem = ElementTree.fromstring(xml_string)
    return elem2json(elem, strip=strip, indent=indent, convert_types=convert_types)

def json2xml(json_string, factory=ElementTree.Element):
    """Convert a JSON string into an XML string."""
    elem = internal2elem(json_loads(json_string), factory)
    return ElementTree.tostring(elem)



# Copyright (c) 2009-2013 Turbulenz Limited
"""
Mesh class used to hold and process vertex streams.

Supports generating NBTs.
"""

import math
import logging
LOG = logging.getLogger('asset')

# pylint: disable=W0403
import vmath
import pointmap
# pylint: enable=W0403

__version__ = '1.1.0'
__dependencies__ = ['pointmap', 'vmath']

#######################################################################################################################

# cos( 1 pi / 8 ) = 0.923879 ~ cos 22.5 degrees
# cos( 2 pi / 8 ) = 0.707106 ~ cos 45   degrees
# cos( 2 pi / 6 ) = 0.5      ~ cos 60   degress
# cos( 3 pi / 8 ) = 0.382683 ~ cos 67.5 degrees
DEFAULT_POSITION_TOLERANCE = 1e-6
DEFAULT_NORMAL_SMOOTH_TOLERANCE = 0.5 # 0.707106
DEFAULT_TANGENT_SPLIT_TOLERANCE = 0.5 # 0.382683
DEFAULT_ZERO_TOLERANCE = 1e-6
DEFAULT_DONT_NORMALIZE_TOLERANCE = 1e-3
DEFAULT_UV_TOLERANCE = 1e-6
DEFAULT_PLANAR_TOLERANCE = 1e-6
DEFAULT_TANGENT_PROJECTION_TOLERANCE = 1e-10
DEFAULT_COLLINEAR_TOLERANCE = 1e-10
DEFAULT_COPLANAR_TOLERANCE = 1e-16
DEFAULT_PLANAR_HULL_VERTEX_THRESHOLD = 5

def similar_positions(major, positions, pos_tol=DEFAULT_POSITION_TOLERANCE):
    """Iterator to return the index of similar positions."""
    # pos_tol currently unused - using vmath default.
    for i, p in enumerate(positions):
        if vmath.v3equal(major, p):
            yield i

#######################################################################################################################

# pylint: disable=R0902
# pylint: disable=R0904
class Mesh():
    """Generate a mesh geometry."""

    class Stream(object):
        """Contains a single stream of vertex attributes."""
        def __init__(self, values, semantic, name, stride, offset):
            """
            ``values`` - array of stream tuples
            ``semantic`` - name of the stream semantic
            ``name`` - human readable name of the stream
            ``stride`` - length of each vertex in the stream
            ``offset`` - offset in the index buffer for this stream
            """
            self.values = values
            self.semantic = semantic
            self.name = name
            self.stride = stride
            self.offset = offset

    def __init__(self, mesh=None):
        # Positions, normals, uvs, skin_indices, skin_weights, primitives
        self.kdtree = None
        if mesh is not None:
            self.positions = mesh.positions[:]
            self.uvs = [ uvs[:] for uvs in mesh.uvs ]
            self.normals = mesh.normals[:]
            self.tangents = mesh.tangents[:]
            self.binormals = mesh.binormals[:]
            self.colors = mesh.colors[:]
            self.skin_indices = mesh.skin_indices[:]
            self.skin_weights = mesh.skin_weights[:]
            self.primitives = mesh.primitives[:]
            self.bbox = mesh.bbox.copy()
        else:
            self.positions = [ ]
            self.uvs = [ [] ]
            self.normals = [ ]
            self.tangents = [ ]
            self.binormals = [ ]
            self.colors = [ ]
            self.skin_indices = [ ]
            self.skin_weights = [ ]
            self.primitives = [ ]
            self.bbox = { }

    ###################################################################################################################

    def set_values(self, values, semantic):
        """Replace the mesh values for a specified semantic."""
        if semantic == 'POSITION':
            self.positions = values
        elif semantic.startswith('TEXCOORD'):
            if semantic == 'TEXCOORD' or semantic == 'TEXCOORD0':
                self.uvs[0] = values
            else:
                index = int(semantic[8:])
                while index >= len(self.uvs):
                    self.uvs.append([])
                self.uvs[index] = values
        elif semantic == 'NORMAL' or semantic == 'NORMAL0':
            self.normals = values
        elif semantic == 'TANGENT':
            self.tangents = values
        elif semantic == 'BINORMAL':
            self.binormals = values
        elif semantic == 'COLOR' or semantic == 'COLOR0':
            self.colors = values
        elif semantic == 'BLENDINDICES':
            self.skin_indices = values
        elif semantic == 'BLENDWEIGHT':
            self.skin_weights = values
        else:
            LOG.warning('Unknown semantic:%s' % semantic)

    def get_values(self, semantic):
        """Retrieve the mesh values for a specified semantic."""
        if semantic == 'POSITION':
            values = self.positions
        elif semantic.startswith('TEXCOORD'):
            if semantic == 'TEXCOORD' or semantic == 'TEXCOORD0':
                values = self.uvs[0]
            else:
                index = int(semantic[8:])
                if index >= len(self.uvs):
                    values = []
                else:
                    values = self.uvs[index]
        elif semantic == 'NORMAL' or semantic == 'NORMAL0':
            values = self.normals
        elif semantic == 'TANGENT':
            values = self.tangents
        elif semantic == 'BINORMAL':
            values = self.binormals
        elif semantic == 'COLOR' or semantic == 'COLOR0':
            values = self.colors
        elif semantic == 'BLENDINDICES':
            values = self.skin_indices
        elif semantic == 'BLENDWEIGHT':
            values = self.skin_weights
        else:
            values = None
            LOG.warning('Unknown semantic:%s' % semantic)
        return values

    ###################################################################################################################

    def transform(self, transform):
        """Transform the vertexes."""
        self.positions = [ vmath.m43transformp(transform, v) for v in self.positions ]
        self.normals = [ vmath.m43transformn(transform, v) for v in self.normals ]
        self.tangents = [ vmath.m43transformn(transform, v) for v in self.tangents ]
        self.binormals = [ vmath.m43transformn(transform, v) for v in self.binormals ]

    def rotate(self, transform):
        """Rotate the vertexes."""
        self.positions = [ vmath.v3mulm33(v, transform) for v in self.positions ]
        self.normals = [ vmath.v3mulm33(v, transform) for v in self.normals ]
        self.tangents = [ vmath.v3mulm33(v, transform) for v in self.tangents ]
        self.binormals = [ vmath.v3mulm33(v, transform) for v in self.binormals ]

    def invert_v_texture_map(self, uvs=None):
        """Invert the v texture mapping."""
        if not uvs:
            uvs = self.uvs[0]
        if len(uvs) > 0:
            if len(uvs[0]) == 2:
                elements = [ ]
                (_, min_v) = uvs[0]
                (_, max_v) = uvs[0]
                for (_, v) in uvs:
                    min_v = min(v, min_v)
                    max_v = max(v, max_v)
                midV = 2 * ((math.ceil(max_v) + math.floor(min_v)) * 0.5)
                for (u, v) in uvs:
                    elements.append( (u, midV - v) )
                self.uvs[0] = elements
            elif len(uvs[0]) == 3:
                elements = [ ]
                (_, min_v, _) = uvs[0]
                (_, max_v, _) = uvs[0]
                for (_, v, _) in uvs:
                    min_v = min(v, min_v)
                    max_v = max(v, max_v)
                midV = 2 * ((math.ceil(max_v) + math.floor(min_v)) * 0.5)
                for (u, v, w) in uvs:
                    elements.append( (u, midV - v, w) )
                self.uvs[0] = elements

    def generate_vertex_with_new_uv(self, pindex, vindex, new_uv):
        """Create a new vertex for a primitive with a new uv."""
        new_vindex = len(self.positions)                        # Get the new vindex
        while len(self.uvs[0]) < new_vindex:                       # Make sure the UVs and Positions are the same size
            self.uvs[0].append( (0, 0) )
        self.positions.append(self.positions[vindex])           # Clone the position
        self.uvs[0].append(new_uv)                                 # Create a new UV
        (ci1, ci2, ci3) = self.primitives[pindex]               # Update the primitive to use the new vindex
        if (ci1 == vindex):
            self.primitives[pindex] = (new_vindex, ci2, ci3)
        elif (ci2 == vindex):
            self.primitives[pindex] = (ci1, new_vindex, ci3)
        elif (ci3 == vindex):
            self.primitives[pindex] = (ci1, ci2, new_vindex)
        else:
            LOG.error("Didn't find vertex:%i used on primitive:%i" % (vindex, pindex))

    def generate_primitives(self, indexes):
        """Generate a list of primitives from a list of indexes."""
        # Make triangles out of each sequence of 3 indices.
        self.primitives = zip(indexes[0::3], indexes[1::3], indexes[2::3])

    def generate_bbox(self):
        """Generate a bounding box for the mesh."""
        # Assumes that the positions have at least one element, or the bbox will be nonsensical
        self.bbox['min'] = (float('inf'), float('inf'), float('inf'))
        self.bbox['max'] = (float('-inf'), float('-inf'), float('-inf'))
        for pos in self.positions:
            self.bbox['min'] = vmath.v3min(self.bbox['min'], pos)
            self.bbox['max'] = vmath.v3max(self.bbox['max'], pos)

    def remove_degenerate_primitives(self, remove_zero_length_edges=True,
                                     edge_length_tol=DEFAULT_POSITION_TOLERANCE):
        """Remove degenerate triangles with duplicated indices and optionally
           zero length edges."""
        def _is_degenerate(prim):
            """ Test if a prim is degenerate """
            (i1, i2, i3) = prim
            if i1 == i2 or i1 == i3 or i2 == i3:
                return True
            elif remove_zero_length_edges:
                (v1, v2, v3) = (self.positions[i1], self.positions[i2], self.positions[i3])
                return (vmath.v3is_zero(vmath.v3sub(v2, v1), edge_length_tol) or
                        vmath.v3is_zero(vmath.v3sub(v3, v1), edge_length_tol) or
                        vmath.v3is_zero(vmath.v3sub(v3, v2), edge_length_tol))
            else:
                return False

        self.primitives = [ prim for prim in self.primitives if not _is_degenerate(prim) ]

    ###################################################################################################################

    def generate_smooth_nbts(self):
        """Helper method to generate and smooth the normals, binormals and tangents."""
        if not len(self.normals):
            # Generate the initial normals if there are none.
            self.generate_normals()
        self.generate_tangents()
        self.normalize_tangents()
        self.smooth_tangents()
        self.generate_normals_from_tangents()
        self.smooth_normals()

    def _generate_normal(self, i1, i2, i3, pos_tol):
        """Generate a normal for 3 indexes."""
        (v1, v2, v3) = (self.positions[i1], self.positions[i2], self.positions[i3])
        e1 = vmath.v3sub(v2, v1)
        e2 = vmath.v3sub(v3, v1)
        e_other = vmath.v3sub(v3, v2)
        if (vmath.v3is_zero(e1, pos_tol) or vmath.v3is_zero(e2, pos_tol) or vmath.v3is_zero(e_other, pos_tol)):
            LOG.warning("%s: Found degenerate primitive:%s with edge length < position tolerance[%g]:[%s,%s,%s]" %
                        ("generate_normals", (i1, i2, i3), pos_tol, e1, e2, e_other))
            return (0, 0, 0)
        return vmath.v3normalize(vmath.v3cross(e1, e2))

    def generate_normals(self, pos_tol=DEFAULT_POSITION_TOLERANCE,
                               dont_norm_tol=DEFAULT_DONT_NORMALIZE_TOLERANCE):
        """Generate a normal per vertex as an average of face normals the primitive is part of."""
        zero = (0, 0, 0)
        self.normals = [zero] * len(self.positions)
        for (i1, i2, i3) in self.primitives:
            n = self._generate_normal(i1, i2, i3, pos_tol)
            self.normals[i1] = vmath.v3add(self.normals[i1], n)
            self.normals[i2] = vmath.v3add(self.normals[i2], n)
            self.normals[i3] = vmath.v3add(self.normals[i3], n)
        for i, n in enumerate(self.normals):
            lsq = vmath.v3lengthsq(n)
            if (lsq > dont_norm_tol): # Ensure normal isn't tiny before normalizing it.
                lr = 1.0 / math.sqrt(lsq)
                self.normals[i] = vmath.v3muls(n, lr)
            else:
                self.normals[i] = zero
                LOG.warning("%s: Found vertex[%i] with normal < normalizable tolerance[%g]:%s" %
                            ("generate_normals", i, dont_norm_tol, n))

    def smooth_normals(self, include_uv_tol=False,
                             root_node=None,
                             pos_tol=DEFAULT_POSITION_TOLERANCE,
                             nor_smooth_tol=DEFAULT_NORMAL_SMOOTH_TOLERANCE,
                             uv_tol=DEFAULT_UV_TOLERANCE):
        """Smooth normals within a certain position range and normal divergence limit."""

        if not root_node:
            if not self.kdtree:
                # Create a kd-tree to optimize the smoothing performance
                self.kdtree = pointmap.build_kdtree(self.positions)
            root_node = self.kdtree

        uvs = self.uvs[0]
        for i, p in enumerate(self.positions):
            original_normal = self.normals[i]
            accumulate_normal = (0, 0, 0)
            accumulated_indexes = [ ]

            # Generate a list of indexes for the positions close to the evaluation vertex.
            if include_uv_tol:
                uv = uvs[i]
                similiar_positions_indexes = root_node.points_within_uv_distance(self.positions, p, pos_tol,
                                                                                 uvs, uv, uv_tol)
            else:
                similiar_positions_indexes = root_node.points_within_distance(self.positions, p, pos_tol)

            for i in similiar_positions_indexes:
                this_normal = self.normals[i]
                if vmath.v3is_similar(this_normal, original_normal, nor_smooth_tol):
                    accumulate_normal = vmath.v3add(accumulate_normal, this_normal)
                    accumulated_indexes.append(i)
            smooth_normal = vmath.v3unitcube_clamp(vmath.v3normalize(accumulate_normal))
            for i in accumulated_indexes:
                self.normals[i] = smooth_normal

    ###################################################################################################################

    def _clone_vertex_with_new_tangents(self, prim_index, vertex_index, tangent, binormal):
        """Create a new vertex clone all attributes. Then set the new tangent and binormal."""
        clone_index = len(self.positions)
        self.positions.append(self.positions[vertex_index])
        self.normals.append(self.normals[vertex_index])
        for uvs in self.uvs:
            if len(uvs) > vertex_index:
                uvs.append(uvs[vertex_index])
        self.tangents.append(tangent)
        self.binormals.append(binormal)
        if len(self.colors) > vertex_index:
            self.colors.append(self.colors[vertex_index])
        if len(self.skin_indices) > vertex_index:
            self.skin_indices.append(self.skin_indices[vertex_index])
        if len(self.skin_weights) > vertex_index:
            self.skin_weights.append(self.skin_weights[vertex_index])
        return clone_index

    def _split_vertex_with_new_tangents(self, vertex_index, prim_index, split_map, tangent, binormal, tan_split_tol_sq):
        """Split the vertex if the tangents are outside of the accumulation tolerance."""

        def update_primitive_vertexes(primitives, prim_index, source_index, target_index):
            """Update the primitive for a cloned set of vertexes."""
            (p1, p2, p3) = primitives[prim_index]
            if p1 == source_index:
                p1 = target_index
            if p2 == source_index:
                p2 = target_index
            if p3 == source_index:
                p3 = target_index
            primitives[prim_index] = (p1, p2, p3)

        def tangents_within_tolerance(t1, b1, t2, b2, tan_split_tol_sq):
            """Test if the tangents and binormals are within tolerance."""
            tangent_within_tolerance = vmath.v3is_within_tolerance(t1, t2, tan_split_tol_sq)
            binormal_within_tolerance = vmath.v3is_within_tolerance(b1, b2, tan_split_tol_sq)
            return tangent_within_tolerance and binormal_within_tolerance

        def potential_accumulation_vertexes(vertex_index, split_map):
            """Iterator to consider all potential accumulation vertexes."""
            yield (vertex_index, vertex_index)
            for (original_index, clone_index) in split_map:
                if original_index == vertex_index:
                    yield (original_index, clone_index)

        for (original_index, index) in potential_accumulation_vertexes(vertex_index, split_map):
            if tangents_within_tolerance(self.tangents[index], self.binormals[index],
                                         tangent, binormal, tan_split_tol_sq):
                self.tangents[index] = vmath.v3add(self.tangents[index], tangent)
                self.binormals[index] = vmath.v3add(self.binormals[index], binormal)
                update_primitive_vertexes(self.primitives, prim_index, original_index, index)
                break
        else:
            # We need to split the vertex to start accumulating a different set of tangents.
            clone_index = self._clone_vertex_with_new_tangents(prim_index, vertex_index, tangent, binormal)
            update_primitive_vertexes(self.primitives, prim_index, vertex_index, clone_index)
            split_map.append( (vertex_index, clone_index) )
            LOG.debug("Splitting vertex:%i --> %i primitive[%i] is now:%s", vertex_index, clone_index, prim_index,
                      self.primitives[prim_index])
            LOG.debug("N:%s B:%s T:%s", self.normals[clone_index], self.binormals[clone_index],
                      self.tangents[clone_index])

    # pylint: disable=R0914
    def _generate_tangents_for_triangle(self, prim, pos_tol, zero_tol):
        """Generate binormals and tangents for a primitive."""
        du = [0, 0, 0]
        dv = [0, 0, 0]
        (i1, i2, i3) = prim                                                                     # Primitive indexes
        (v1, v2, v3) = (self.positions[i1], self.positions[i2], self.positions[i3])             # Vertex positions
        uvs = self.uvs[0]
        (uv1, uv2, uv3) = (uvs[i1], uvs[i2], uvs[i3])                                           # Vertex UVs
        (e21, e31, e32) = (vmath.v3sub(v2, v1), vmath.v3sub(v3, v1), vmath.v3sub(v3, v2))       # Generate edges
        # Ignore degenerates
        if (vmath.v3is_zero(e21, pos_tol) or vmath.v3is_zero(e31, pos_tol) or vmath.v3is_zero(e32, pos_tol)):
            LOG.warning("%s: Found degenerate triangle:%s" % ("_generate_tangents_for_triangle", (i1, i2, i3)))
        else:
            # Calculate tangent and binormal
            edge1 = [e21[0], uv2[0] - uv1[0], uv2[1] - uv1[1]]
            edge2 = [e31[0], uv3[0] - uv1[0], uv3[1] - uv1[1]]
            cp = vmath.v3cross(edge1, edge2)
            if not vmath.iszero(cp[0], zero_tol):
                du[0] = -cp[1] / cp[0]
                dv[0] = -cp[2] / cp[0]
            edge1[0] = e21[1] # y, s, t
            edge2[0] = e31[1]
            cp = vmath.v3cross(edge1, edge2)
            if not vmath.iszero(cp[0], zero_tol):
                du[1] = -cp[1] / cp[0]
                dv[1] = -cp[2] / cp[0]
            edge1[0] = e21[2] # z, s, t
            edge2[0] = e31[2]
            cp = vmath.v3cross(edge1, edge2)
            if not vmath.iszero(cp[0], zero_tol):
                du[2] = -cp[1]/cp[0]
                dv[2] = -cp[2]/cp[0]
        return (du, dv)
    # pylint: enable=R0914

    def generate_tangents(self, pos_tol=DEFAULT_POSITION_TOLERANCE,
                                zero_tol=DEFAULT_ZERO_TOLERANCE,
                                tan_split_tol=DEFAULT_TANGENT_SPLIT_TOLERANCE):
        """Generate a NBT per vertex."""
        if 0 == len(self.uvs[0]): # We can't generate nbts without uvs
            LOG.debug("Can't generate nbts without uvs:%i"% (len(self.uvs[0])))
            return
        num_vertices = len(self.positions)
        self.tangents = [ (0, 0, 0) ] * num_vertices
        self.binormals = [ (0, 0, 0) ] * num_vertices

        # Split map for recording pairs of integers that represent which vertexes have been split.
        partition_vertices = True
        split_map = [ ]
        tan_split_tol_sq = (tan_split_tol * tan_split_tol)

        for prim_index, prim in enumerate(self.primitives):
            (tangent, binormal) = self._generate_tangents_for_triangle(prim, pos_tol, zero_tol)
            (i1, i2, i3) = prim
            if partition_vertices:
                self._split_vertex_with_new_tangents(i1, prim_index, split_map, tangent, binormal, tan_split_tol_sq)
                self._split_vertex_with_new_tangents(i2, prim_index, split_map, tangent, binormal, tan_split_tol_sq)
                self._split_vertex_with_new_tangents(i3, prim_index, split_map, tangent, binormal, tan_split_tol_sq)
            else:
                # Accumulate tangent and binormal
                self.tangents[i1] = vmath.v3add(self.tangents[i1], tangent)
                self.tangents[i2] = vmath.v3add(self.tangents[i2], tangent)
                self.tangents[i3] = vmath.v3add(self.tangents[i3], tangent)
                self.binormals[i1] = vmath.v3add(self.binormals[i1], binormal)
                self.binormals[i2] = vmath.v3add(self.binormals[i2], binormal)
                self.binormals[i3] = vmath.v3add(self.binormals[i3], binormal)

    def normalize_tangents(self, dont_norm_tol=DEFAULT_DONT_NORMALIZE_TOLERANCE):
        """Normalize and clamp the new tangents and binormals."""
        zero = (0, 0, 0)

        tangents = [ ]
        for i, t in enumerate(self.tangents):
            if (vmath.v3lengthsq(t) > dont_norm_tol): # Ensure the tangent isn't tiny before normalizing it.
                tangents.append(vmath.v3unitcube_clamp(vmath.v3normalize(t)))
            else:
                LOG.warning("%s: Found vertex[%i] with tangent < normalizable tolerance[%g]:%s" %
                            ("normalize_tangents", i, dont_norm_tol, t))
                tangents.append(zero)
        self.tangents = tangents

        binormals = [ ]
        for i, b in enumerate(self.binormals):
            if (vmath.v3lengthsq(b) > dont_norm_tol):
                binormals.append(vmath.v3unitcube_clamp(vmath.v3normalize(b)))
            else:
                LOG.warning("%s: Found vertex[%i] with binormal < normalizable tolerance[%g]:%s" %
                            ("normalize_tangents", i, dont_norm_tol, b))
                binormals.append(zero)
        self.binormals = binormals

    def smooth_tangents(self, include_uv_tol=False,
                              root_node=None,
                              pos_tol=DEFAULT_POSITION_TOLERANCE,
                              nor_smooth_tol=DEFAULT_NORMAL_SMOOTH_TOLERANCE,
                              uv_tol=DEFAULT_UV_TOLERANCE):
        """Smooth the tangents of vertices with similar positions."""
        def tangents_are_similar(t1, t2, b1, b2, nor_smooth_tol):
            """Test if both the tangents and binormals are similar."""
            tangent_similar = vmath.v3is_similar(t1, t2, nor_smooth_tol)
            binormal_similar = vmath.v3is_similar(b1, b2, nor_smooth_tol)
            return tangent_similar and binormal_similar

        if not root_node:
            if not self.kdtree:
                # Create a kd-tree to optimize the smoothing performance
                self.kdtree = pointmap.build_kdtree(self.positions)
            root_node = self.kdtree

        uvs = self.uvs[0]
        for i, p in enumerate(self.positions):
            original_tangent = self.tangents[i]
            original_binormal = self.binormals[i]
            accumulate_tangent = (0, 0, 0)
            accumulate_binormal = (0, 0, 0)
            accumulated_indexes = [ ]

            # Generate a list of indexes for the positions close to the evaluation vertex.
            if include_uv_tol:
                uv = uvs[i]
                similiar_positions_indexes = root_node.points_within_uv_distance(self.positions, p, pos_tol,
                                                                                 uvs, uv, uv_tol)
            else:
                similiar_positions_indexes = root_node.points_within_distance(self.positions, p, pos_tol)

            for i in similiar_positions_indexes:
                this_tangent = self.tangents[i]
                this_binormal = self.binormals[i]
                if tangents_are_similar(this_tangent, original_tangent,
                                        this_binormal, original_binormal, nor_smooth_tol):
                    accumulate_tangent = vmath.v3add(accumulate_tangent, this_tangent)
                    accumulate_binormal = vmath.v3add(accumulate_binormal, this_binormal)
                    accumulated_indexes.append(i)
            smooth_tangent = vmath.v3unitcube_clamp(vmath.v3normalize(accumulate_tangent))
            smooth_binormal = vmath.v3unitcube_clamp(vmath.v3normalize(accumulate_binormal))
            for i in accumulated_indexes:
                self.tangents[i] = smooth_tangent
                self.binormals[i] = smooth_binormal

    def generate_normals_from_tangents(self, zero_tol=DEFAULT_ZERO_TOLERANCE,
                                             dont_norm_tol=DEFAULT_DONT_NORMALIZE_TOLERANCE):
        """Create a new normal from the tangent and binormals."""
        if not len(self.tangents) or not len(self.binormals): # We can't generate normals without nbts
            LOG.debug("Can't generate normals from nbts without tangets:%i and binormals:%i"
                      % (len(self.tangents), len(self.binormals)))
            return
        num_vertices = len(self.normals)
        assert(num_vertices == len(self.tangents))
        assert(num_vertices == len(self.binormals))
        # Regenerate the vertex normals from the new tangents and binormals
        for i in range(num_vertices):
            normal = self.normals[i]
            cp = vmath.v3cross(self.tangents[i], self.binormals[i])
            # Keep vertex normal if the tangent and the binormal are paralel
            if (vmath.v3lengthsq(cp) > dont_norm_tol):
                cp = vmath.v3normalize(cp)
                # Keep vertex normal if new normal is *somehow* in the primitive plane
                cosangle = vmath.v3dot(cp, normal)
                if not vmath.iszero(cosangle, zero_tol):
                    if cosangle < 0:
                        self.normals[i] = vmath.v3neg(cp)
                    else:
                        self.normals[i] = cp

    def flip_primitives(self):
        """Change winding order"""
        self.primitives = [ (i1, i3, i2) for (i1, i2, i3) in self.primitives ]

    def mirror_in(self, axis="x", flip=True):
        """Flip geometry in axis."""
        if axis == "x":
            self.positions = [ (-x, y, z) for (x, y, z) in self.positions ]
            self.normals = [ (-x, y, z) for (x, y, z) in self.normals ]
        elif axis == "y":
            self.positions = [ (x, -y, z) for (x, y, z) in self.positions ]
            self.normals = [ (x, -y, z) for (x, y, z) in self.normals ]
        elif axis == "z":
            self.positions = [ (x, y, -z) for (x, y, z) in self.positions ]
            self.normals = [ (x, y, -z) for (x, y, z) in self.normals ]
        if flip:
            self.flip_primitives()

    ###################################################################################################################

    def remove_redundant_vertexes(self):
        """Remove redundant vertex indexes from the element streams if unused by the primitives."""
        mapping = { }
        new_index = 0
        for (i1, i2, i3) in self.primitives:
            if i1 not in mapping:
                mapping[i1] = new_index
                new_index += 1
            if i2 not in mapping:
                mapping[i2] = new_index
                new_index += 1
            if i3 not in mapping:
                mapping[i3] = new_index
                new_index += 1
        old_index = len(self.positions)
        if old_index != new_index:
            LOG.info("Remapping:remapping vertexes from %i to %i", old_index, new_index)

        def __remap_stream(source, size, mapping):
            """Remap vertex attribute stream."""
            if len(source) > 0:
                target = [0] * size
                for k, v in mapping.items():
                    target[v] = source[k]
                return target
            return [ ]

        self.positions = __remap_stream(self.positions, new_index, mapping)
        for uvs in self.uvs:
            uvs[:] = __remap_stream(uvs, new_index, mapping)
        self.normals = __remap_stream(self.normals, new_index, mapping)
        self.tangents = __remap_stream(self.tangents, new_index, mapping)
        self.binormals = __remap_stream(self.binormals, new_index, mapping)
        self.colors = __remap_stream(self.colors, new_index, mapping)
        self.skin_indices = __remap_stream(self.skin_indices, new_index, mapping)
        self.skin_weights = __remap_stream(self.skin_weights, new_index, mapping)

        primitives = [ ]
        for (i1, i2, i3) in self.primitives:
            primitives.append( (mapping[i1], mapping[i2], mapping[i3]) )
        self.primitives = primitives

    ###################################################################################################################

    def stitch_vertices(self):
        """Combine equal vertices together, adjusting indices of primitives where appropriate
           Any other vertex data like normals, tangents, colors are ignored"""

        num_points = len(self.positions)
        points = sorted(enumerate(self.positions), key=lambda (_, x): x)

        mapping = [0] * num_points
        new_index = -1
        prev_p = None
        for (index, p) in points:
            if p != prev_p:
                new_index += 1
                prev_p = p
            mapping[index] = new_index

        self.primitives = [(mapping[i1], mapping[i2], mapping[i3]) for (i1, i2, i3) in self.primitives]
        new_positions = [0] * (new_index + 1)
        for (i, to) in enumerate(mapping):
            new_positions[to] = self.positions[i]
        self.positions = new_positions

    ###################################################################################################################

    def is_convex(self, positions=None, primitives=None):
        """Check if a mesh is convex by validating no vertices lie in front of the planes defined by its faces."""
        positions = positions or self.positions
        primitives = primitives or self.primitives
        for (i1, i2, i3) in primitives:
            v1 = positions[i1]
            v2 = positions[i2]
            v3 = positions[i3]

            edge1 = vmath.v3sub(v1, v3)
            edge2 = vmath.v3sub(v2, v3)
            normal = vmath.v3normalize(vmath.v3cross(edge1, edge2))

            for p in positions:
                dist = vmath.v3dot(vmath.v3sub(p, v1), normal)
                if dist > vmath.PRECISION:
                    return False
        return True

    def simply_closed(self, primitives=None):
        """Determine if a connected mesh is closed defined by triangle indexes
           in sense that it defines the boundary of a closed region of space without any
           possible dangling triangles on boundary.

           We assume that the triangle mesh does not have any ugly self intersections."""
        primitives = primitives or self.primitives
        # We do this by counting the number of triangles on each triangle edge
        # Specifically check that this value is exactly 2 for every edge.1
        edges = { }
        def _inc(a, b):
            """Increment edge count for vertex indices a, b.
               Return True if edge count has exceeded 2"""
            if a > b:
                return _inc(b, a)
            if (a, b) in edges:
                edges[(a, b)] += 1
            else:
                edges[(a, b)] = 1
            return edges[(a, b)] > 2

        for (i1, i2, i3) in primitives:
            if _inc(i1, i2):
                return False
            if _inc(i2, i3):
                return False
            if _inc(i3, i1):
                return False

        for (_, v) in edges.items():
            if v != 2:
                return False

        return True

    def is_planar(self, positions=None, tolerance=DEFAULT_PLANAR_TOLERANCE):
        """Determine if mesh is planar; that all vertices lie in the same plane.
           If positions argument is not supplied, then self.positions is used"""
        positions = positions or self.positions
        if len(positions) <= 3:
            return True

        p0 = positions[0]
        edge1 = vmath.v3sub(positions[1], p0)
        edge2 = vmath.v3sub(positions[2], p0)
        normal = vmath.v3normalize(vmath.v3cross(edge1, edge2))

        for p in positions:
            distance = vmath.v3dot(vmath.v3sub(p, p0), normal)
            if (distance * distance) > tolerance:
                return False

        return True

    def is_convex_planar(self, positions=None):
        """Determine if a planar mesh is convex; taking virtual edge normals into account"""
        positions = positions or self.positions
        if len(positions) <= 3:
            return True

        p0 = positions[0]
        edge1 = vmath.v3sub(positions[1], p0)
        edge2 = vmath.v3sub(positions[2], p0)
        normal = vmath.v3normalize(vmath.v3cross(edge1, edge2))

        for (i, p) in enumerate(positions):
            j = (i + 1) if (i < len(positions) - 1) else 0
            q = positions[j]

            edge3 = vmath.v3sub(q, p)
            edge_normal = vmath.v3cross(normal, edge3)

            for w in positions:
                if vmath.v3dot(vmath.v3sub(w, p), edge_normal) < -vmath.PRECISION:
                    return False

        return True

    def connected_components(self):
        """Determine connected components of mesh, returning list of set of vertices and primitives."""
        # Perform this algorithm with a disjoint set forest.
        # Initialise components: [index, parent_index, rank, component]
        components = [[i, 0] for i in range(len(self.positions))]

        def _find(x):
            """Find root note in disjoint set forest, compressing path to root."""
            if components[x][0] == x:
                return x
            else:
                root = x
                stack = [ ]
                while components[root][0] != root:
                    stack.append(root)
                    root = components[root][0]
                for y in stack:
                    components[y][0] = root
                return root

        def _unify(x, y):
            """Unify two components in disjoint set forest by rank."""
            x_root = _find(x)
            y_root = _find(y)
            if x_root != y_root:
                xc = components[x_root]
                yc = components[y_root]
                if xc[1] < yc[1]:
                    xc[0] = y_root
                elif xc[1] > yc[1]:
                    yc[0] = x_root
                else:
                    yc[0] = x_root
                    xc[1] += 1

        # Unify components based on adjacency information inferred through shared vertices in primitives
        for (i1, i2, i3) in self.primitives:
            _unify(i1, i2)
            _unify(i2, i3)

        # Return list of all components associated with each root.
        ret = [ ]
        for c in [y for y in range(len(self.positions)) if _find(y) == y]:
            m = Mesh()
            m.positions = self.positions[:]
            m.primitives = [(i1, i2, i3) for (i1, i2, i3) in self.primitives if _find(i1) == c]

            m.remove_redundant_vertexes()
            ret.append((m.positions, m.primitives))

        return ret

    # pylint: disable=R0914
    def make_planar_convex_hull(self, positions=None, tangent_tolerance=DEFAULT_TANGENT_PROJECTION_TOLERANCE):
        """Convert set of co-planar positions into a minimal set of positions required to form their convex
           hull, together with a set of primitives representing one side of the hulls' surface as a
           new Mesh"""
        positions = positions or self.positions
        # Use a 2D Graham Scan with projections of positions onto their maximal plane.
        # Time complexity: O(nh) for n positions and h out positions.

        # Determine maximal plane for projection.
        edge1 = vmath.v3sub(positions[1], positions[0])
        edge2 = vmath.v3sub(positions[2], positions[0])
        (n0, _, n2) = n = vmath.v3cross(edge1, edge2)

        # compute plane tangents.
        # epsilon chosen with experiment
        if (n0 * n0) + (n2 * n2) < tangent_tolerance:
            t = (1, 0, 0)
        else:
            t = (-n2, 0, n0)
        u = vmath.v3cross(n, t)

        # Project to tangents
        projs = [(vmath.v3dot(p, t), vmath.v3dot(p, u)) for p in positions]

        # Find first vertex on hull as minimal lexicographically ordered projection
        i0 = 0
        minp = projs[0]
        for i in range(1, len(projs)):
            if projs[i] < minp:
                i0 = i
                minp = projs[i]

        # Map from old vertex indices to new index for those vertices used by hull
        outv = { i0: 0 }
        new_index = 1

        # List of output triangles.
        outtriangles = [ ]
        fsti = i0
        (p0x, p0y) = minp
        while True:
            i1 = -1
            for i in range(len(projs)):
                if i == i0:
                    continue

                (px, py) = projs[i]
                plsq = ((px - p0x) * (px - p0x)) + ((py - p0y) * (py - p0y))
                if i1 == -1:
                    i1 = i
                    maxp = (px, py)
                    maxplsq = plsq
                    continue

                # If this is not the first vertex tested, determine if new vertex makes
                # A right turn looking in direction of edge, or is further in same direction.
                (qx, qy) = maxp
                turn = ((qx - p0x) * (py - p0y)) - ((qy - p0y) * (px - p0x))
                if turn < 0 or (turn == 0 and plsq > maxplsq):
                    i1 = i
                    maxp = (px, py)
                    maxplsq = plsq

            # Append i1 vertex to hull
            if i1 in outv:
                break

            outv[i1] = new_index
            new_index += 1

            # Form triangle (fsti, i0, i1)
            # If i0 != fsti
            if i0 != fsti:
                outtriangles.append((fsti, i0, i1))

            i0 = i1
            (p0x, p0y) = projs[i1]

        # Compute output hull.
        mesh = Mesh()
        mesh.positions = [0] * len(outv.items())
        for i, j in outv.items():
            mesh.positions[j] = positions[i]
        mesh.primitives = [(outv[i1], outv[i2], outv[i3]) for (i1, i2, i3) in outtriangles]

        return mesh
    # pylint: enable=R0914

    # pylint: disable=R0914
    def make_convex_hull(self, positions=None, collinear_tolerance=DEFAULT_COLLINEAR_TOLERANCE,
                               coplanar_tolerance=DEFAULT_COPLANAR_TOLERANCE):
        """Convert set of positions into a minimal set of positions required to form their convex hull
           Together with a set of primitives representing a triangulation of the hull's surface as a
           new Mesh"""
        positions = positions or self.positions
        # Use a 3D generalisation of a Graham Scan to facilitate a triangulation of the hull in generation.
        # Time complexity: O(nh) for n positions, and h out-positions.

        # Find first vertex on hull as minimal lexicographically ordered position
        i0 = 0
        minp = positions[0]
        for i in range(1, len(positions)):
            if positions[i] < minp:
                i0 = i
                minp = positions[i]

        # Find second vertex by performing a 2D graham scan step on the xy-plane projections of positions.
        i1 = -1
        (cos1, lsq1) = (-2, 0) # will always be overriden as cos(theta) > -2
        (p0x, p0y, _) = minp
        for i in range(len(positions)):
            if i == i0:
                continue

            (px, py, _) = positions[i]
            dx = px - p0x
            dy = py - p0y
            lsq = (dx * dx) + (dy * dy)
            if lsq == 0:
                if i1 == -1:
                    i1 = i
                continue

            cos = dy / math.sqrt(lsq)
            if cos > cos1 or (cos == cos1 and lsq > lsq1):
                cos1 = cos
                lsq1 = lsq
                i1 = i

        # Dictionary of visited edges to avoid duplicates
        # List of open edges to be visited by graham scan
        closedset = set()
        openset = [ (i0, i1), (i1, i0) ]

        # Mapping from old vertex index to new index for those vertices used by hull.
        outv = { i0: 0, i1: 1 }
        new_index = 2

        # Output triangles for hull
        outtriangles = [ ]

        while len(openset) > 0:
            (i0, i1) = openset.pop()
            if (i0, i1) in closedset:
                continue

            # Find next vertex on hull to form triangle with.
            i2 = -1

            p0 = positions[i0]
            edge = vmath.v3sub(positions[i1], p0)
            isq = 1.0 / vmath.v3lengthsq(edge)

            for i in range(len(positions)):
                if i == i0 or i == i1:
                    continue

                p = positions[i]
                # Find closest point on line containing the edge to determine vector to p
                # Perpendicular to edge, this is not necessary for computing the turn
                # since the value of 'turn' computed is actually the same whether we do this
                # or not, however it is needed to be able to sort equal turn vertices
                # by distance.
                t = vmath.v3dot(vmath.v3sub(p, p0), edge) * isq
                pedge = vmath.v3sub(p, vmath.v3add(p0, vmath.v3muls(edge, t)))

                # Ignore vertex if |pedge| = 0, thus ignoring vertices on the edge itself
                # And so avoid generating degenerate triangles.
                #
                # epsilon chosen by experiment
                plsq = vmath.v3lengthsq(pedge)
                if plsq <= collinear_tolerance:
                    continue

                if i2 == -1:
                    i2 = i
                    maxpedge = pedge
                    maxplsq = plsq
                    maxt = t
                    continue

                # If this is not the first vertex tested, determine if new vertex makes
                # A right turn looking in direction of edge, or is further in same direction.
                #
                # We require a special case when pedge, and maxpedge are coplanar with edge
                # As the computed turn will be 0 and we must check if the cross product
                # Is facing into the hull or outside to determine left/right instead.
                axis = vmath.v3cross(pedge, maxpedge)
                coplanar = vmath.v3dot(pedge, vmath.v3cross(edge, maxpedge))
                # epsilon chosen by experiment
                if (coplanar * coplanar) <= coplanar_tolerance:
                    # Special case for coplanar pedge, maxpedge, edge
                    #
                    # if edges are in same direction, base on distance.
                    if vmath.v3dot(pedge, maxpedge) >= 0:
                        if plsq > maxplsq or (plsq == maxplsq and t > maxt):
                            i2 = i
                            maxpedge = pedge
                            maxplsq = plsq
                            maxt = t
                    else:
                        axis = vmath.v3cross(vmath.v3sub(p, p0), edge)
                        # Check if axis points into the hull.
                        internal = True
                        for p in positions:
                            if vmath.v3dot(axis, vmath.v3sub(p, p0)) < 0:
                                internal = False
                                break

                        if internal:
                            i2 = i
                            maxpedge = pedge
                            maxplsq = plsq
                            maxt = t
                else:
                    turn = vmath.v3dot(axis, edge)
                    # epsilon chosen by experiment
                    if turn < 0 or (turn <= collinear_tolerance and plsq > maxplsq):
                        i2 = i
                        maxpedge = pedge
                        maxplsq = plsq
                        maxt = t

            # Append i2 vertex to hull
            if i2 not in outv:
                outv[i2] = new_index
                new_index += 1

            # Form triangle iff no edge is closed.
            if ((i0, i1) not in closedset and
                (i1, i2) not in closedset and
                (i2, i0) not in closedset):

                outtriangles.append((i0, i1, i2))
                # Mark visited edges. Open new edges.
                closedset.add((i0, i1))
                closedset.add((i1, i2))
                closedset.add((i2, i0))

                openset.append((i2, i1))
                openset.append((i0, i2))

        # cnt does not 'need' to be len(positions) for convex hull to have succeeded
        #   but numerical issues with not using say fixed point means that we cannot
        #   be sure of success if it is not equal.
        # Obvious side effect is input mesh must already be a convex hull with no
        #   unnecessary vertices.
        cnt = len(outv.items())
        if cnt != len(positions):
            return None

        # Compute output mesh.
        mesh = Mesh()
        mesh.positions = [0] * cnt
        for (i, j) in outv.items():
            mesh.positions[j] = positions[i]
        mesh.primitives = [(outv[i1], outv[i2], outv[i3]) for (i1, i2, i3) in outtriangles]

        # Ensure algorithm has not failed!
        if not mesh.is_convex() or not mesh.simply_closed():
            return None

        return mesh
    # pylint: enable=R0914

    def extend_mesh(self, positions, primitives):
        """Extend mesh with extra set of positions and primitives defiend relative to positions"""
        offset = len(self.positions)
        self.positions.extend(positions)
        self.primitives.extend([(i1 + offset, i2 + offset, i3 + offset) for (i1, i2, i3) in primitives])

    def convex_hulls(self, max_components=-1, allow_non_hulls=False,
                           planar_vertex_count=DEFAULT_PLANAR_HULL_VERTEX_THRESHOLD):
        """Split triangle mesh into set of unconnected convex hulls.

           If max_components != -1, then None will be returned should the number
           of connected components exceed this value.

           If allow_non_hulls is False, and any component of the mesh was not able
           to be converted then None will be returned.

           No other vertex data is assumed to exist, and mesh is permitted to be
           mutated.

           The return value is a tuple ([Mesh], Mesh) for the list of convex hulls
           computed, and an additional mesh representing the remainder of the mesh
           which could not be converted (If allow_non_hulls is False, then this
           additional mesh will always be None, otherwise it may still be None
           if all of the mesh was able to be converted)."""
        self.stitch_vertices()
        self.remove_degenerate_primitives()
        self.remove_redundant_vertexes()

        components = self.connected_components()
        if max_components != -1 and len(components) > max_components:
            return None

        if allow_non_hulls:
            triangles = Mesh()
            triangles.positions = [ ]
            triangles.primitives = [ ]

        ret = [ ]
        for (vertices, primitives) in components:
            convex = self.is_convex(vertices, primitives)
            closed = self.simply_closed(primitives)
            planar = self.is_planar(vertices)

            if convex and planar:
                if self.is_convex_planar(vertices) and len(vertices) >= planar_vertex_count:
                    print "Converted to planar convex hull!"
                    ret.append(self.make_planar_convex_hull(vertices))
                else:
                    if allow_non_hulls:
                        triangles.extend_mesh(vertices, primitives)
                    else:
                        return None

            elif convex and closed:
                mesh = self.make_convex_hull(vertices)
                if mesh == None:
                    if allow_non_hulls:
                        print "Failed to turn convex closed mesh into convex hull!"
                        triangles.extend_mesh(vertices, primitives)
                    else:
                        return None
                else:
                    # Ensure that convex hull can be re-computed correctly as this will be performed
                    # By WebGL physics device.
                    if mesh.make_convex_hull() == None:
                        if allow_non_hulls:
                            print "Convex hull failed to be re-computed!"
                            triangles.extend_mesh(vertices, primitives)
                        else:
                            return None
                    else:
                        print "Converted to convex hull!"
                        ret.append(mesh)

            else:
                # Cannot convert component to a convex hull.
                if allow_non_hulls:
                    triangles.extend_mesh(vertices, primitives)
                else:
                    return None

        if len(triangles.positions) == 0:
            triangles = None

        return (ret, triangles)

# pylint: enable=R0902
# pylint: enable=R0904

#######################################################################################################################

# pylint: disable=C0111
def __generate_test_square(json_asset):

    def __generate_square_t():
        mesh = Mesh()
        mesh.positions.extend( [ (0, 0, 0), (1, 0, 0), (1, 1, 1), (0, 1, 1), (2, 0, 0), (2, 1, 1) ] )
        mesh.uvs[0].extend( [ (0, 0), (1, 0), (1, 1), (0, 1), (0, 0), (0, 1) ] )
        indexes = [ ]
        indexes.extend( [ 0, 1, 2, 0, 2, 3, 1, 4, 5, 1, 5, 2 ] )
        return (mesh, indexes)

    def __generate_square_b():
        mesh = Mesh()
        mesh.positions.extend( [ (0, 2, 2), (1, 2, 2), (1, 3, 3), (0, 3, 3), (2, 2, 2), (2, 3, 3) ] )
        mesh.uvs[0].extend( [ (0, 0), (0, 1), (1, 1), (1, 0), (0, 0), (1, 0) ] )
        indexes = [ ]
        indexes.extend( [ 0, 1, 2, 0, 2, 3, 1, 4, 5, 1, 5, 2 ] )
        return (mesh, indexes)

    def __generate_split_square_t():
        mesh = Mesh()
        mesh.positions.extend( [ (0, 4, 4), (1, 4, 4), (1, 5, 5), (0, 5, 5), (2, 4, 4), (2, 5, 5),
                                 (1, 4, 4), (1, 5, 5) ] )
        mesh.uvs[0].extend( [ (0, 0), (1, 0), (1, 1), (0, 1), (0, 0), (0, 1),
                              (1, 0), (1, 1) ] )
        indexes = [ ]
        indexes.extend( [ 0, 1, 2, 0, 2, 3, 6, 4, 5, 6, 5, 7 ] )
        return (mesh, indexes)

    def __generate_split_square_b():
        mesh = Mesh()
        mesh.positions.extend( [ (0, 6, 6), (1, 6, 6), (1, 7, 7), (0, 7, 7), (2, 6, 6), (2, 7, 7),
                                 (1, 6, 6), (1, 7, 7) ] )
        mesh.uvs[0].extend( [ (0, 0), (0, 1), (1, 1), (1, 0), (0, 0), (1, 0),
                              (0, 1), (1, 1) ] )
        indexes = [ ]
        indexes.extend( [ 0, 1, 2, 0, 2, 3, 6, 4, 5, 6, 5, 7 ] )
        return (mesh, indexes)

    def __generate_control_square_t():
        r2 = 1 / math.sqrt(2)
        mesh = Mesh()
        mesh.positions.extend( [ (0, 8, 8), (1, 8, 8), (1, 9, 9), (0, 9, 9),
                                 (1, 8, 8), (2, 8, 8), (2, 9, 9), (1, 9, 9) ] )
        mesh.uvs[0].extend( [ (0, 0), (1, 0), (1, 1), (0, 1),
                              (1, 0), (0, 0), (0, 1), (1, 1) ] )
        mesh.tangents.extend( [ (1, 0, 0), (1, 0, 0), (1, 0, 0), (1, 0, 0),
                                (-1, 0, 0), (-1, 0, 0), (-1, 0, 0), (-1, 0, 0) ] )
        mesh.binormals.extend( [ (0, r2, r2), (0, r2, r2), (0, r2, r2), (0, r2, r2),
                                 (0, r2, r2), (0, r2, r2), (0, r2, r2), (0, r2, r2) ] )
        indexes = [ ]
        indexes.extend( [ 0, 1, 2, 0, 2, 3, 4, 5, 6, 4, 6, 7 ] )
        return (mesh, indexes)

    def __generate_control_square_b():
        r2 = 1 / math.sqrt(2)
        mesh = Mesh()
        mesh.positions.extend( [ (0, 10, 10), (1, 10, 10), (1, 11, 11), (0, 11, 11),
                                 (1, 10, 10), (2, 10, 10), (2, 11, 11), (1, 11, 11) ] )
        mesh.uvs[0].extend( [ (0, 0), (0, 1), (1, 1), (1, 0),
                              (0, 1), (0, 0), (1, 0), (1, 1) ] )
        mesh.tangents.extend( [ (0, r2, r2), (0, r2, r2), (0, r2, r2), (0, r2, r2),
                                (0, r2, r2), (0, r2, r2), (0, r2, r2), (0, r2, r2) ] )
        mesh.binormals.extend( [ (1, 0, 0), (1, 0, 0), (1, 0, 0), (1, 0, 0),
                                 (-1, 0, 0), (-1, 0, 0), (-1, 0, 0), (-1, 0, 0) ] )
        indexes = [ ]
        indexes.extend( [ 0, 1, 2, 0, 2, 3, 4, 5, 6, 4, 6, 7 ] )
        return (mesh, indexes)

    def __generate_square((a, i), (s, n), j):
        a.generate_primitives(i)
        a.generate_normals()
        if len(a.tangents) == 0:
            a.generate_smooth_nbts()
        j.attach_shape(s)
        n = NodeName(n)
        j.attach_node(n)
        j.attach_node_shape_instance(n, s, s, m)
        j.attach_positions(a.positions, s)
        j.attach_nbts(a.normals, a.tangents, a.binormals, s)
        j.attach_uvs(a.uvs[0], s)
        j.attach_surface(a.primitives, JsonAsset.SurfaceTriangles, s)

    m = 'material-0'
    e = 'effect-0'
    json_asset.attach_effect(e, 'normalmap')
    json_asset.attach_material(m, e)
    json_asset.attach_texture(m, 'diffuse', '/assets/checker.png')
    json_asset.attach_texture(m, 'normal_map', '/assets/monkey.png')
    __generate_square(__generate_square_t(), ('shape-0', 'node-0'), json_asset)
    __generate_square(__generate_square_b(), ('shape-1', 'node-1'), json_asset)
    __generate_square(__generate_split_square_t(), ('shape-2', 'node-2'), json_asset)
    __generate_square(__generate_split_square_b(), ('shape-3', 'node-3'), json_asset)
    __generate_square(__generate_control_square_t(), ('shape-4', 'node-4'), json_asset)
    __generate_square(__generate_control_square_b(), ('shape-5', 'node-5'), json_asset)

def __generate_test_cube():
    def add_quad_face(v, m, primitives):
        """Append a two triangle quad to the mesh."""
        offset = len(m.positions)
        m.positions.extend( [ v[0], v[1], v[2], v[3] ] )
        m.uvs[0].extend( [ (1, 1), (0, 1), (0, 0), (1, 0) ] )
        primitives.extend( [offset, offset + 2, offset + 1, offset, offset + 3, offset + 2] )

    indexes = [ ]
    mesh = Mesh()
    v = [ (-1, -1, -1), (-1, -1,  1), (-1,  1,  1), (-1,  1, -1),
          ( 1, -1, -1), ( 1, -1,  1), ( 1,  1,  1), ( 1,  1, -1) ]
    add_quad_face( [ v[0], v[1], v[2], v[3] ], mesh, indexes)
    add_quad_face( [ v[4], v[0], v[3], v[7] ], mesh, indexes)
    add_quad_face( [ v[5], v[4], v[7], v[6] ], mesh, indexes)
    add_quad_face( [ v[1], v[5], v[6], v[2] ], mesh, indexes)
    add_quad_face( [ v[7], v[3], v[2], v[6] ], mesh, indexes)
    add_quad_face( [ v[1], v[0], v[4], v[5] ], mesh, indexes)

    mesh.generate_primitives(indexes)
    return mesh

# pylint: enable=C0111

if __name__ == "__main__":
    # pylint: disable=W0403
    from asset2json import JsonAsset
    from node import NodeName
    # pylint: enable=W0403
    logging.basicConfig(level=logging.INFO)

    J = JsonAsset()
    __generate_test_square(J)
    J.clean()
    JSON = J.json_to_string()
    with open("mesh.json", 'w') as output:
        output.write(JSON)
    print JSON
    J.log_metrics()



# Copyright (c) 2009-2011,2013 Turbulenz Limited

"""
Collection of utility functions for handling subprocess execution.
"""

import datetime
import subprocess

__version__ = '1.0.0'

class SubProc:
    """Encapsulation for running subprocesses, capturing the output and processing to return in a response."""
    def __init__(self, command, cwd=None):
        self.command = command
        self.cwd = cwd
        self.retcode = 0
        self.time_delta = datetime.timedelta()
        self.stdout_report, self.stderr_report = ('','')

    def update_command(self, command, cwd=None, env=None):
        """Set the command to execute and optionally the path to execute form."""
        self.command = command
        self.cwd = cwd

    def time_popen(self):
        """Time a subprocess command and return process retcode. This method will block until the process completes."""
        time_start = datetime.datetime.now()

        proc = subprocess.Popen(self.command, stdout=subprocess.PIPE, stderr=subprocess.PIPE, cwd=self.cwd)
        stdout_report, stderr_report = proc.communicate()
        self.retcode = proc.wait()

        time_end = datetime.datetime.now()
        time_delta = time_end - time_start

        self.time_delta += time_delta
        self.stdout_report += stdout_report
        self.stderr_report += stderr_report

        return self.retcode

    def command_str(self):
        """Generate the command string."""
        return ' '.join(self.command)



# Copyright (c) 2009-2011,2013 Turbulenz Limited
"""
Turbulenz framework for creating browser based games.
"""

__author__ = "Turbulenz Limited"
__copyright__   = "Copyright (c) 2009-2011,2013 Turbulenz Limited"

__version__ = '1.0.1'



#!/usr/bin/env python
# Copyright (c) 2010-2013 Turbulenz Limited

import logging

from re import sub
from optparse import OptionParser, OptionGroup, TitledHelpFormatter

from turbulenz_tools.utils.xml_json import xml2json
from turbulenz_tools.tools.stdtool import simple_options

__version__ = '1.0.0'
__dependencies__ = ['turbulenz_tools.utils.xml_json']

LOG = logging.getLogger(__name__)

def _parser():
    usage = "usage: %prog [options] -i source.xml -o output.json"
    description = "Convert XML assets into a structured JSON asset."

    parser = OptionParser(description=description, usage=usage, formatter=TitledHelpFormatter())
    parser.add_option("--version", action="store_true", dest="output_version", default=False,
                      help="output version number")
    parser.add_option("-v", "--verbose", action="store_true", dest="verbose", default=False, help="verbose output")
    parser.add_option("-s", "--silent", action="store_true", dest="silent", default=False, help="silent running")
    parser.add_option("-m", "--metrics", action="store_true", dest="metrics", default=False,
                      help="output asset metrics")

    parser.add_option("-i", "--input", action="store", dest="input", help="input XML file to process")
    parser.add_option("-o", "--output", action="store", dest="output", help="output JSON file to process")

    group = OptionGroup(parser, "Asset Generation Options")
    group.add_option("-j", "--json-indent", action="store", dest="json_indent", type="int", default=0, metavar="SIZE",
                     help="json output pretty printing indent size, defaults to 0")
    group.add_option("-n", "--namespace", action="store_true", dest="namespace", default=False,
                     help="maintain XML xmlns namespace in JSON asset keys.")
    group.add_option("-c", "--convert-types", action="store_true", dest="convert_types", default=False,
                     help="attempt to convert values to ints, floats and lists.")

    parser.add_option_group(group)

    return parser

def main():
    (options, args_, parser_) = simple_options(_parser, __version__, __dependencies__)

    try:
        with open(options.input) as xml_file:
            xml_string = xml_file.read()

            # At the moment there doesn't seem to be an obvious way to extract the xmlns from the asset.
            # For now, we'll attempt to just remove it before transforming it into a Python object.

            # <COLLADA xmlns="http://www.collada.org/2005/11/COLLADASchema" version="1.4.1">
            #   ==>
            # <COLLADA version="1.4.1">
            if options.namespace is False:
                xml_string = sub(' xmlns="[^"]*"', '', xml_string)

            json_string = xml2json(xml_string, indent=options.json_indent, convert_types=options.convert_types)
            if options.output:
                with open(options.output, 'w') as target:
                    target.write(json_string)
                    target.write('\n')
            else:
                print json_string

    except IOError as e:
        LOG.error(e)
        return e.errno
    except Exception as e:
        LOG.critical('Unexpected exception: %s' % e)
        return 1

if __name__ == "__main__":
    exit(main())



#!/usr/bin/python
# Copyright (c) 2011-2013 Turbulenz Limited
"""
Convert Bitmap Font Generator data (.fnt) files into a Turbulenz JSON asset.
http://www.angelcode.com/products/bmfont/
"""

import re
import sys
import logging

from optparse import OptionParser, OptionGroup, TitledHelpFormatter

# pylint: disable=W0403
from stdtool import standard_output_version, standard_json_out
from asset2json import JsonAsset
# pylint: enable=W0403

__version__ = '2.0.0'
__dependencies__ = ['asset2json']


LOG = logging.getLogger('asset')


#######################################################################################################################

class Bmfont2json():
    """Parse a .fnt file and generate a Turbulenz JSON geometry asset."""

    bold_re = re.compile(r'bold=(\d+)')
    italic_re = re.compile(r'italic=(\d+)')
    page_width_re = re.compile(r'scaleW=(\d+)')
    page_height_re = re.compile(r'scaleH=(\d+)')
    line_height_re = re.compile(r'lineHeight=(\d+)')
    base_re = re.compile(r'base=(\d+)')
    num_pages_re = re.compile(r'pages=(\d+)')
    page_re = re.compile(r'page\s+id=(\d+)\s+file="(\S+)"')
    num_chars_re = re.compile(r'chars\s+count=(\d+)')
    num_kernings_re = re.compile(r'kernings\s+count=(\d+)')
    kerning_re = re.compile(r'kerning\s+first=(\d+)\s+second=(\d+)\s+amount=([-+]?\d+)')
    # pylint: disable=C0301
    char_re = re.compile(r'char\s+id=(\d+)\s+x=(\d+)\s+y=(\d+)\s+width=(\d+)\s+height=(\d+)\s+xoffset=([-+]?\d+)\s+yoffset=([-+]?\d+)\s+xadvance=([-+]?\d+)\s+page=(\d+)')
    # pylint: enable=C0301

    def __init__(self, texture_prefix):
        self.bold = 0
        self.italic = 0
        self.glyphs = { }
        self.page_width = 0
        self.page_height = 0
        self.baseline = 0
        self.line_height = 0
        self.num_glyphs = 0
        self.min_glyph_index = 256
        self.texture_pages = []
        self.kernings = { }
        self.texture_prefix = texture_prefix

    def __read_page(self, line):
        found = self.page_re.match(line)
        if not found:
            raise Exception('Page information espected, found: ' + line)

        page_index = int(found.group(1))
        page_file = self.texture_prefix + found.group(2)
        self.texture_pages[page_index] = page_file

        LOG.info("texture page: %s", page_file)

    def __read_char(self, line):
        """Parse one glyph descriptor."""
        found = self.char_re.match(line)
        if not found:
            raise Exception('Char information espected, found: ' + line)

        i = int(found.group(1))
        x = float(found.group(2))
        y = float(found.group(3))
        width = int(found.group(4))
        height = int(found.group(5))
        xoffset = int(found.group(6))
        yoffset = int(found.group(7))
        xadvance = int(found.group(8))
        page = int(found.group(9))

        self.glyphs[i] = {
            'width': width,
            'height': height,
            'awidth': xadvance,
            'xoffset': xoffset,
            'yoffset': yoffset,
            'left': x / self.page_width,
            'top': y / self.page_height,
            'right': (x + width) / self.page_width,
            'bottom': (y + height) / self.page_height,
            'page': page
        }

        if self.min_glyph_index > i:
            self.min_glyph_index = i

    def __read_kerning(self, line):
        """Parse one kerning descriptor."""
        found = self.kerning_re.match(line)
        if not found:
            raise Exception('Kerning information espected, found: ' + line)

        first = int(found.group(1))
        second = int(found.group(2))
        amount = int(found.group(3))

        kerning = self.kernings.get(first, None)
        if kerning is None:
            self.kernings[first] = {second: amount}
        else:
            kerning[second] = amount

#######################################################################################################################

    def parse(self, f):
        """Parse a .fnt file stream."""

        line = f.readline()
        line = line.strip()

        if line.startswith('info '):
            found = self.bold_re.search(line)
            if found:
                self.bold = int(found.group(1))

            found = self.italic_re.search(line)
            if found:
                self.italic = int(found.group(1))

            LOG.info("bold: %d", self.bold)
            LOG.info("italic: %d", self.italic)

            line = f.readline()
            line = line.strip()

        # Common
        if not line.startswith('common '):
            raise Exception('Common information espected, found: ' + line)

        found = self.page_width_re.search(line)
        if not found:
            raise Exception('ScaleW espected, found: ' + line)

        self.page_width = int(found.group(1))

        LOG.info("page width: %d", self.page_width)

        found = self.page_height_re.search(line)
        if not found:
            raise Exception('ScaleH espected, found: ' + line)

        self.page_height = int(found.group(1))

        LOG.info("page height: %d", self.page_height)

        found = self.line_height_re.search(line)
        if not found:
            raise Exception('Line Height espected, found: ' + line)

        self.line_height = int(found.group(1))

        LOG.info("line height: %d", self.line_height)

        found = self.base_re.search(line)
        if not found:
            raise Exception('Base espected, found: ' + line)

        self.baseline = int(found.group(1))

        LOG.info("baseline: %d", self.baseline)

        found = self.num_pages_re.search(line)
        if not found:
            raise Exception('Num pages espected, found: ' + line)

        num_pages = int(found.group(1))
        self.texture_pages = [None] * num_pages

        LOG.info("num texture pages: %d", num_pages)

        # Pages
        line = f.readline()
        line = line.strip()
        while line.startswith('page'):
            self.__read_page(line)
            line = f.readline()
            line = line.strip()

        found = self.num_chars_re.search(line)
        if not found:
            raise Exception('Chars count espected, found: ' + line)

        self.num_glyphs = int(found.group(1))
        if self.num_glyphs <= 0:
            raise Exception('No glyphs found!')

        LOG.info("num glyphs: %d", self.num_glyphs)

        line = f.readline()
        line = line.strip()
        while line.startswith('char'):
            self.__read_char(line)
            line = f.readline()
            line = line.strip()

        # Kernings
        found = self.num_kernings_re.search(line)
        if found:
            num_kernings = int(found.group(1))
            if num_kernings > 0:
                line = f.readline()
                line = line.strip()
                while line.startswith('kerning'):
                    self.__read_kerning(line)
                    line = f.readline()
                    line = line.strip()
        else:
            num_kernings = 0

        LOG.info("num kernings: %d", num_kernings)


    def get_definitions(self, filename):
        """Return a fixed asset object."""
        filename = filename.replace('\\', '/')
        asset = {
            'version': 1,
            'bitmapfontlayouts' : {
                filename: {
                    'pagewidth': self.page_width,
                    'pageheight': self.page_height,
                    'baseline': self.baseline,
                    'lineheight': self.line_height,
                    'numglyphs': self.num_glyphs,
                    'minglyphindex': self.min_glyph_index,
                    'glyphs': self.glyphs,
                    'pages': self.texture_pages
                }
            }
        }
        if self.kernings:
            asset['bitmapfontlayouts'][filename]['kernings'] = self.kernings
        if self.bold:
            asset['bitmapfontlayouts'][filename]['bold'] = True
        if self.italic:
            asset['bitmapfontlayouts'][filename]['italic'] = True
        return asset


#######################################################################################################################

def bmfont2json_parser(description, epilog=None):
    """Standard set of parser options."""
    parser = OptionParser(description=description, epilog=epilog,
                          formatter=TitledHelpFormatter())

    parser.add_option("--version", action="store_true", dest="output_version", default=False,
                      help="output version number")
    parser.add_option("-v", "--verbose", action="store_true", dest="verbose", default=False, help="verbose outout")
    parser.add_option("-s", "--silent", action="store_true", dest="silent", default=False, help="silent running")
    parser.add_option("-m", "--metrics", action="store_true", dest="metrics", default=False,
                      help="output asset metrics")
    parser.add_option("--log", action="store", dest="output_log", default=None, help="write log to file")

    group = OptionGroup(parser, "Asset Generation Options")
    group.add_option("-j", "--json_indent", action="store", dest="json_indent", type="int", default=0, metavar="SIZE",
                     help="json output pretty printing indent size, defaults to 0")
    parser.add_option_group(group)

    group = OptionGroup(parser, "Asset Location Options")
    group.add_option("-p", "--prefix", action="store", dest="texture_prefix", default="textures/", metavar="URL",
                     help="texture URL to prefix to all texture references")
    group.add_option("-a", "--assets", action="store", dest="asset_root", default=".", metavar="PATH",
                     help="PATH of the asset root")
    parser.add_option_group(group)

    group = OptionGroup(parser, "File Options")
    group.add_option("-i", "--input", action="store", dest="input", default=None, metavar="FILE",
                     help="source FILE to process")
    group.add_option("-o", "--output", action="store", dest="output", default="default.json", metavar="FILE",
                     help="output FILE to write to")
    parser.add_option_group(group)

    return parser

def parse(input_filename="default.fontdat", output_filename="default.json", texture_prefix="", asset_root=".",
          options=None):
    """Untility function to convert an .fnt file into a JSON file."""
    with open(input_filename, 'rb') as source:
        asset = Bmfont2json(texture_prefix)
        try:
            asset.parse(source)
            asset_name = input_filename
            if input_filename.startswith(asset_root):
                asset_name = asset_name[(len(asset_root) + 1):-8]
            json_asset = JsonAsset(definitions=asset.get_definitions(asset_name))
            standard_json_out(json_asset, output_filename, options)
            return json_asset
        # pylint: disable=W0703
        except Exception as e:
            LOG.error(str(e))
        # pylint: enable=W0703

def main():
    description = ("Convert Bitmap Font Generator data (.fnt) files into a Turbulenz JSON asset.\n" +
                   "http://www.angelcode.com/products/bmfont/")

    parser = bmfont2json_parser(description)

    (options, args_) = parser.parse_args()

    if options.output_version:
        standard_output_version(__version__, __dependencies__, options.output)
        return

    if options.input is None:
        parser.print_help()
        return

    if options.silent:
        level = logging.CRITICAL
    elif options.verbose or options.metrics:
        level = logging.INFO
    else:
        level = logging.WARNING
    logging.basicConfig(level=level, stream=sys.stdout)

    LOG.info("input: %s", options.input)
    LOG.info("output: %s", options.output)

    if options.texture_prefix != '':
        options.texture_prefix = options.texture_prefix.replace('\\', '/')
        if options.texture_prefix[-1] != '/':
            options.texture_prefix = options.texture_prefix + '/'
        LOG.info("texture URL prefix: %s", options.texture_prefix)

    if options.asset_root != '.':
        LOG.info("root: %s", options.asset_root)

    parse(options.input,
          options.output,
          options.texture_prefix,
          options.asset_root,
          options)

if __name__ == "__main__":
    exit(main())



#!/usr/bin/env python
# Copyright (c) 2012-2013 Turbulenz Limited

# Catch "Exception"
# pylint:disable=W0703

from logging import getLogger
from os.path import splitext, basename

from optparse import OptionParser, TitledHelpFormatter

from turbulenz_tools.utils.dependencies import find_dependencies
from turbulenz_tools.utils.dependencies import find_file_in_dirs
from turbulenz_tools.utils.profiler import Profiler

from turbulenz_tools.tools.templates import env_create
from turbulenz_tools.tools.templates import env_load_template
from turbulenz_tools.tools.templates import env_load_templates

from turbulenz_tools.tools.appcodegen import render_js
from turbulenz_tools.tools.appcodegen import context_from_options
from turbulenz_tools.tools.appcodegen import default_add_code
from turbulenz_tools.tools.appcodegen import inject_js_from_options
from turbulenz_tools.tools.appcodegen import default_parser_options
from turbulenz_tools.tools.appcodegen import DEFAULT_HTML_TEMPLATE
from turbulenz_tools.tools.appcodegen import output_dependency_info

from turbulenz_tools.tools.toolsexception import ToolsException
from turbulenz_tools.tools.stdtool import simple_options

__version__ = '1.8.0'
__dependencies__ = ['turbulenz_tools.utils.dependencies', 'turbulenz_tools.tools.appcodegen']

LOG = getLogger(__name__)

############################################################

def _parser():
    parser = OptionParser(description='Generate HTML files from .html and '
                          '.js files. Any options not recognised are '
                          'assumed to be input files.',
                          usage="usage: %prog [options] <input files>",
                          formatter=TitledHelpFormatter())

    default_parser_options(parser)

    parser.add_option("-C", "--code", action="store", dest="codefile",
                      help="release file to be called by the HTML. Does not "
                      "need to exist yet. (release and canvas modes only)")

    # Mode one of [ 'plugin', 'plugin-debug', 'canvas', 'canvas-debug' ]
    parser.add_option("-m", "--mode", action="store", dest="mode",
                      default='plugin-debug',
                      help="build mode: canvas, canvas-debug, plugin, "
                      "plugin-debug (default)")

    parser.add_option("-D", "--dump-default-template", action="store_true",
                      dest="dump_default_template", default=False,
                      help="output the default template to file")

    return parser

############################################################

# TODO : Move into utils
def check_input(input_files):
    """
    Divide up a list of input files into .js and .html files
    """
    js_files = []
    html_files = []
    for f in input_files:
        ext = splitext(f)[1]
        if ext in [ '.js', '.jsinc' ]:
            js_files.append(f)
        elif ext in [ '.html', '.htm' ]:
            html_files.append(f)
        else:
            LOG.error("unrecognised file type: %s" % f)
            exit(1)

    return (js_files, html_files)

def load_html_template(env, input_html):
    if 1 == len(input_html):
        return env_load_template(env, input_html[0])

    return env.from_string(DEFAULT_HTML_TEMPLATE)

def dump_default_template(outfile_name):
    if outfile_name is None:
        outfile_name = 'default_template.html'

    with open(outfile_name, "wb") as f:
        f.write(DEFAULT_HTML_TEMPLATE)
    LOG.info("Default template written to: %s" % outfile_name)
    return 0

def html_dump_dependencies(env, options, input_js, input_html):
    """
    Dump the dependencies of the html file being output
    """

    # For html, dependencies are:
    # - dev: html template deps, top-level js files
    # - release: html template deps
    # - canvas_dev: html template deps, top-level js files
    # - canvas: html template deps

    outfile_name = options.dependency_file
    if outfile_name is None:
        LOG.error("No dependency output file specified")
        return 1

    # Collect html dependencies (if there are html files available)

    if 1 == len(input_html):
        try:
            deps = find_dependencies(input_html[0], options.templatedirs, env,
                                     [ 'default' ])
        except Exception, e:
            raise ToolsException("dependency error: %s" % str(e))
    else:
        deps = []

    # Collect js dependencies if necessary

    if options.mode in [ 'plugin-debug', 'canvas-debug' ]:
        deps += [ find_file_in_dirs(js, options.templatedirs) for js in input_js ]

    # Write dependency info

    output_dependency_info(outfile_name, options.output, deps)

    return 0

def html_generate(env, options, input_js, input_html):
    """
    Generate html based on the templates and build mode.
    """

    # - dev, canvas_dev:
    #     render top-level js files into a temporary file
    #     collect the .js files that need to be included
    #     setup includes, startup code and the js render result into variables
    #     render html template
    #
    # - release, canvas:
    #     need to know name of output js file
    #     setup startup code to point to .tzjs or .js file
    #     render html template

    # Load templates (using default html template if not specified)

    Profiler.start('load_templates')

    template_html = load_html_template(env, input_html)
    if template_html is None:
        LOG.error("failed to load file %s from template dirs" % input_html[0])
        exit(1)

    # Get context

    if len(input_js) > 0:
        title = input_js[0]
    elif options.codefile:
        title = options.codefile
    elif len(input_html) > 0:
        title = input_html[0]
    else:
        title = "Unknown"
    title = splitext(basename(title))[0]

    context = context_from_options(options, title)

    Profiler.stop('load_templates')
    Profiler.start('code_gen')

    # In development modes, render the JS code that needs embedding

    rendered_js = ""
    inc_js = []

    if options.mode in [ 'plugin-debug', 'canvas-debug' ]:
        inject_js = inject_js_from_options(options)

        Profiler.start('load_js_templates')
        templates_js = env_load_templates(env, input_js)
        Profiler.stop('load_js_templates')

        (rendered_js, inc_js) = render_js(context, options, templates_js,
                                          inject_js)

    # Add the HTML and JS code into the tz_* variables

    default_add_code(options, context, rendered_js, inc_js)

    Profiler.stop('code_gen')
    Profiler.start('html_render')

    # Render the template and write it out

    try:
        res = template_html.render(context)
    except Exception, e:
        raise ToolsException("Error in '%s': %s %s" \
                                 % (input_html, e.__class__.__name__, str(e)))

    try:
        with open(options.output, "wb") as f:
            f.write(res)
    except IOError:
        raise ToolsException("failed to create file: %s" % options.output)

    Profiler.stop('html_render')

    return 0

############################################################

def main():

    (options, args, parser) = simple_options(_parser, __version__,
                                             __dependencies__, input_required=False)

    Profiler.start('main')
    Profiler.start('startup')

    input_files = args

    # Check that if dump-default-template is set then output and exit

    if options.dump_default_template:
        exit(dump_default_template(options.output))
    elif 0 == len(args):
        LOG.error('No input files specified')
        parser.print_help()
        exit(1)

    LOG.info("options: %s" % options)
    LOG.info("args: %s" % args)
    LOG.info("parser: %s" % parser)
    LOG.info("templatedirs: %s" % options.templatedirs)

    if options.output is None:
        LOG.error("no output file specified (required in dependency mode)")
        parser.print_help()
        exit(1)

    # Check mode

    if options.mode not in [ 'plugin-debug', 'plugin', 'canvas-debug', 'canvas' ]:
        LOG.error('Unrecognised mode: %s' % options.mode)
        parser.print_help()
        exit(1)

    # Check a release source name is given if mode is one of release
    # or canvas

    if options.mode in [ 'plugin', 'canvas' ] and \
            not options.dependency and \
            not options.codefile:
        LOG.error('Missing code file name.  Use --code to specify.')
        parser.print_usage()
        exit(1)

    # Check input files and split them into (ordered) js and html

    (input_js, input_html) = check_input(input_files)

    LOG.info("js files: %s" % input_js)
    LOG.info("html files: %s" % input_html)

    # In debug and canvas-debug we need a .js input file

    if 0 == len(input_js):
        if options.mode in [ 'debug', 'canvas-debug' ]:
            LOG.error('Missing input .js file')
            parser.print_usage()
            exit(1)
    if 1 < len(input_html):
        LOG.error('Multiple html files specified: %s' % input_html)
        exit(1)

    # Create a jinja2 env

    env = env_create(options, DEFAULT_HTML_TEMPLATE)

    Profiler.stop('startup')
    Profiler.start('run')

    # Execute

    retval = 1
    try:

        if options.dependency:
            LOG.info("generating dependencies")
            retval = html_dump_dependencies(env, options, input_js, input_html)
            LOG.info("done generating dependencies")

        else:
            retval = html_generate(env, options, input_js, input_html)

    except ToolsException, e:
        #traceback.print_exc()
        LOG.error("%s" % str(e))

    Profiler.stop('run')
    Profiler.stop('main')
    Profiler.dump_data()

    return retval

############################################################

if __name__ == "__main__":
    exit(main())



# Copyright (c) 2009-2013 Turbulenz Limited
"""
Utilities to simplify building a standard translater.
"""

import sys
import logging
LOG = logging.getLogger('asset')

from os.path import basename as path_basename, exists as path_exists
from simplejson import load as json_load
from optparse import OptionParser, OptionGroup, TitledHelpFormatter

# pylint: disable=W0403
from asset2json import JsonAsset
from turbulenz_tools.utils.json_utils import merge_dictionaries
# pylint: enable=W0403

#######################################################################################################################


#######################################################################################################################

def standard_output_version(version, dependencies, output_file):
    main_module_name = path_basename(sys.argv[0])
    version_string = None
    if dependencies:
        deps = { }

        def get_dependencies_set(this_module_name, deps_list):
            for module_name in deps_list:
                if module_name not in deps:
                    m = None
                    try:
                        m = __import__(module_name, globals(), locals(),
                                       ['__version__', '__dependencies__'])
                    except ImportError:
                        print "Failed to import %s, listed in dependencies " \
                            "for %s" % (module_name, this_module_name)
                        exit(1)
                    else:
                        # Test is the module actually has a version attribute
                        try:
                            version_ = m.__version__
                        except AttributeError as e:
                            print 'No __version__ attribute for tool %s' \
                                % m.__name__
                            print ' >> %s' % str(e)
                        else:
                            deps[module_name] = m

                    if m is not None:
                        try:
                            get_dependencies_set(module_name,
                                                 m.__dependencies__)
                        except AttributeError:
                            pass

        get_dependencies_set(main_module_name, dependencies)

        module_names = deps.keys()
        module_names.sort()

        module_list = ', '.join(['%s %s' % (deps[m].__name__, deps[m].__version__) for m in module_names])
        version_string = '%s %s (%s)' % (main_module_name, version, module_list)
    else:
        version_string = '%s %s' % (main_module_name, version)

    # If we are given an output file, write the versions info there if
    # either:
    #   the file doesn't exist already, or
    #   the file contains different data
    # If we are given no output file, just write to stdout.

    print version_string
    if output_file is not None:
        if path_exists(output_file):
            with open(output_file, "rb") as f:
                old_version = f.read()
            if old_version == version_string:
                return
        with open(output_file, "wb") as f:
            f.write(version_string)

def standard_include(infiles):
    """Load and merge all the ``infiles``."""
    if infiles:
        definitions = { }
        for infile in infiles:
            if path_exists(infile):
                with open(infile, 'r') as infile_file:
                    infile_json = json_load(infile_file)
                    definitions = merge_dictionaries(infile_json, definitions)
            else:
                LOG.error('Missing file: %s' % infile)
        return JsonAsset(definitions=definitions)
    else:
        return JsonAsset()
    return None

def standard_parser(description, epilog=None, per_file_options=True):
    """Standard set of parser options."""
    parser = OptionParser(description=description, epilog=epilog,
                          formatter=TitledHelpFormatter())
    parser.add_option("--version", action="store_true", dest="output_version",
                      default=False, help="output version number to output "
                      "file, or stdout if no output file is given")
    parser.add_option("-v", "--verbose", action="store_true", dest="verbose",
                      default=False, help="verbose outout")
    parser.add_option("-s", "--silent", action="store_true", dest="silent",
                      default=False, help="silent running")
    if per_file_options:
        parser.add_option("-m", "--metrics", action="store_true",
                          dest="metrics", default=False, help="output asset "
                          "metrics")
    parser.add_option("--log", action="store", dest="output_log", default=None,
                      help="write log to file")

    group = OptionGroup(parser, "Asset Generation Options")
    group.add_option("-j", "--json_indent", action="store", dest="json_indent",
                     type="int", default=0, metavar="SIZE",
                     help="json output pretty printing indent size, defaults "
                     "to 0")

    # TODO - Asset Generation Options currently disabled
    #
    #group.add_option("-6", "--base64-encoding", action="store_true", dest="b64_encoding", default=False,
    #                 help=("encode long float and int attributes in base64, defaults to disabled %s" %
    #                        "- [ currently unsupported ]"))
    #group.add_option("-c", "--force-collision", action="store_true", dest="force_collision", default=False,
    #                 help="force collision generation - [ currently unsupported ]")
    #group.add_option("-r", "--force-render", action="store_true", dest="force_render", default=False,
    #                 help="force rendering generation - [ currently unsupported ]")

    group.add_option("--keep-unused-images", action="store_true", dest="keep_unused_images", default=False,
                     help="keep images with no references to them")

    group.add_option("-I", "--include-type", action="append", dest="include_types", default=None, metavar="TYPE",
                     help="only include objects of class TYPE in export.")
    group.add_option("-E", "--exclude-type", action="append", dest="exclude_types", default=None, metavar="TYPE",
                     help="exclude objects of class TYPE from export. "
                          "Classes currently supported for include and exclude: "
                          "geometries, nodes, animations, images, effects, materials, lights, "
                          "physicsmaterials, physicsmodels and physicsnodes. "
                          "CAUTION using these options can create incomplete assets which require fixup at runtime. ")
    parser.add_option_group(group)

    group = OptionGroup(parser, "Asset Location Options")
    group.add_option("-u", "--url", action="store", dest="asset_url", default="", metavar="URL",
                     help="asset URL to prefix to all asset references")
    group.add_option("-a", "--assets", action="store", dest="asset_root", default=".", metavar="PATH",
                     help="PATH of the asset root")
    group.add_option("-d", "--definitions", action="append", dest="definitions", default=None, metavar="JSON_FILE",
                     help="definition JSON_FILE to include in build, this option can be used repeatedly for multiple "
                          "files")
    parser.add_option_group(group)

    if per_file_options:
        group = OptionGroup(parser, "File Options")
        group.add_option("-i", "--input", action="store", dest="input", default=None, metavar="FILE",
                         help="source FILE to process")
        group.add_option("-o", "--output", action="store", dest="output", default="default.json", metavar="FILE",
                         help="output FILE to write to")
        parser.add_option_group(group)

    # TODO - Database Options are currently disabled
    #
    #group = OptionGroup(parser, "Database Options")
    #group.add_option("-A", "--authority", action="store", dest="authority", default=None,
    #                 metavar="HOST:PORT",
    #                 help=("Authority of the database in the form HOST:PORT. %s" %s
    #                       "If undefined, database export is disabled."))
    #group.add_option("-D", "--database", action="store", dest="database", default="default",
    #                 metavar="NAME", help="NAME of the document database")
    #group.add_option("-P", "--put-post", action="store_true", dest="put_post", default=False,
    #                 help="put or post the asset to the authority database")
    #group.add_option("-O", "--document", action="store", dest="document", default="default.asset",
    #                 metavar="NAME", help="NAME of the document")
    #parser.add_option_group(group)

    return parser

def standard_main(parse, version, description, dependencies, parser = None):
    """Provide a consistent wrapper for standalone translation.
       When parser is not supplied, standard_parser(description) is used."""

    parser = parser or standard_parser(description)
    (options, args_) = parser.parse_args()

    if options.output_version:
        standard_output_version(version, dependencies, options.output)
        return

    if options.input is None:
        parser.print_help()
        return

    if options.silent:
        logging.basicConfig(level=logging.CRITICAL, stream=sys.stdout)
    elif options.verbose or options.metrics:
        logging.basicConfig(level=logging.INFO, stream=sys.stdout)
    else:
        logging.basicConfig(level=logging.WARNING, stream=sys.stdout)

    LOG.info("input: %s" % (options.input))
    LOG.info("output: %s" % (options.output))

    if options.asset_url != '':
        LOG.info("url: %s" % (options.asset_url))

    if options.asset_root != '.':
        LOG.info("root: %s" % (options.asset_root))

    if options.definitions:
        for inc in options.definitions:
            LOG.info("inc: %s" % (inc))

    # TODO - Database Options are currently disabled
    #
    #if options.put_post:
    #    LOG.info("## authority: %s" % (options.authority))

    json_asset_ = parse(options.input, options.output,
                        options.asset_url, options.asset_root, options.definitions,
                        options)

    # TODO - Database Options are currently disabled
    #
    #if options.put_post:
    #    try:
    #        from couchdbkit import Server
    #        from couchdbkit.resource import PreconditionFailed
    #        database_supported = True
    #    except ImportError:
    #        database_supported = False
    #
    #    if database_supported and json_asset and options.put_post:
    #        server_uri = 'http://' + options.authority + '/'
    #        server = Server(uri=server_uri)
    #        try:
    #            database = server.get_or_create_db(options.database)
    #        except PreconditionFailed:
    #            database = server[options.database]
    #        if database.doc_exist(options.document):
    #            pass
    #        else:
    #            database[options.document] = json_asset.asset

def standard_json_out(json_asset, output_filename, options=None):
    """Provide a consistent output of the JSON assets."""

    indent = 0
    if options is not None:
        indent = options.json_indent

    metrics = False
    if options is not None:
        metrics = options.metrics

    json_asset.clean()
    if metrics:
        json_asset.log_metrics()

    with open(output_filename, 'w') as target:
        json_asset.json_to_file(target, True, indent)
        target.write('\n')

#######################################################################################################################

def simple_options(parser_fn, version, dependencies, input_required=True):
    parser = parser_fn()
    (options, args) = parser.parse_args()

    if options.output_version:
        standard_output_version(version, dependencies, options.output)
        exit(0)

    if input_required:
        # Not all tools have an input file, so we print help for no args as well.
        try:
            if options.input is None:
                print "ERROR: no input files specified"
                parser.print_help()
                exit(1)
        except AttributeError:
            if len(args) == 0:
                print "ERROR: no input files specified"
                parser.print_help()
                exit(1)

    # Not all tools have a metrics option.
    try:
        metrics = options.metrics
    except AttributeError:
        metrics = False

    if options.silent:
        logging.basicConfig(level=logging.CRITICAL)
    elif options.verbose or metrics:
        logging.basicConfig(level=logging.INFO)
    else:
        logging.basicConfig(level=logging.WARNING)

    return (options, args, parser)



# Copyright (c) 2009-2011,2013 Turbulenz Limited
"""
Utility for manipluating Nodes.
"""

__version__ = '1.0.0'

class NodeName(object):
    """Node class to consistent name ."""

    def __init__(self, name):
        self.name = name
        self.path = [ ]

    def add_parent(self, parent):
        """Add the name for a parent node."""
        self.path.append(parent)

    def add_parents(self, parents):
        """Add a list of names for the parent nodes."""
        self.path.extend(parents)

    def add_parent_node(self, parent_node):
        """Add a parent node as a parent."""
        self.add_parents(parent_node.hierarchy_names())
        return self

    def add_path(self, path):
        """Add a path for the parent nodes which is split by '/'."""
        self.path = path.split('/')

    def leaf_name(self):
        """Return the leaf node name."""
        return self.name

    def hierarchy_names(self):
        """Return a list of all the node names."""
        return self.path + [self.name]

    def __repr__(self):
        return self.__str__()

    def __str__(self):
        if len(self.path) > 0:
            return '/'.join(self.path) + '/' + self.name
        else:
            return self.name



# Copyright (c) 2010-2011,2013 Turbulenz Limited
"""
Utilities.
"""

from glob import glob as standard_glob

def glob(pattern):
    """Slash consistent globbing."""
    return [ f.replace('\\', '/') for f in standard_glob(pattern) ]



# Copyright (c) 2010-2011,2013 Turbulenz Limited

import re
import logging

from HTMLParser import HTMLParser

# pylint: disable=W0403
from jsmin import jsmin
# pylint: enable=W0403

LOG = logging.getLogger(__name__)

# pylint: disable=R0904
class HTMLMinifier(HTMLParser):
    """An HTML minifier."""

    REMOVE_WHITESPACE = re.compile(r'\s{2,}').sub

    def __init__(self, output, compact_script=True):
        """output: This callback function will be called when there is data to output.
        A good candidate to use is sys.stdout.write."""
        HTMLParser.__init__(self)
        self.output = output
        self.compact_script = compact_script
        self.pre_count = 0
        self.inside_script = False

    def error(self, message):
        LOG.warning('Warning: %s' % message)

    def handle_starttag(self, tag, attributes):
        if 'pre' == tag:
            self.pre_count += 1
        elif 'script' == tag:
            script_type = None
            for (key, value) in attributes:
                if key == 'type':
                    script_type = value
                    break
            if script_type != 'text/html':
                self.inside_script = True
        # This is no longer required as the controller can now signal the middleware to not compact the response.
        #elif 'html' == tag:
        #    # If the request doesn't contain an html tag - we can't assume it isn't inserted within a pre tag (this
        #    # happens in the disassembler.) So we only 'enable' white space removal if we see the html tag.
        #    self.pre_count = 0

        data = self.REMOVE_WHITESPACE(' ', self.get_starttag_text())
        self.output(data)

    def handle_startendtag(self, tag, attributes):
        #self.handle_starttag(tag, attributes)
        #self.handle_endtag(tag)
        data = self.REMOVE_WHITESPACE(' ', self.get_starttag_text())
        self.output(data)

    def handle_endtag(self, tag):
        if 'pre' == tag:
            self.pre_count -= 1
        elif 'script' == tag:
            self.inside_script = False
        self.output('</%s>' % tag)

    def handle_data(self, data):
        if self.inside_script:
            if self.compact_script:
                data = jsmin(data)
        elif self.pre_count == 0:
            data = self.REMOVE_WHITESPACE(' ', data)
            if data == ' ':
                return
        self.output(data)

    def handle_charref(self, name):
        self.output('&#%s;' % name)

    def handle_entityref(self, name):
        self.output('&%s;' % name)

    def handle_comment(self, data):
        return

    def handle_decl(self, data):
        self.output('<!%s>' % data)
        return

    def handle_pi(self, data):
        return
# pylint: enable=R0904



# Copyright (c) 2009-2011,2013 Turbulenz Limited
"""Utility module to colour highlight the output form tools."""

import sys
import re

# pylint: disable=R0903
class ColouredWriterBase():
    """Colour th Django server output."""
    colors = {  'endc': '\033[0m',              # black
                'fail': '\033[91m',             # red
                'okgreen': '\033[32m',          # green
                'unknown': '\033[33m',          # yellow
                'okblue': '\033[34m',           # blue
                'warning': '\033[95m',          # magenta
                'build': '\033[97m\033[40m' }   # white

    status_code_text = {
            100: 'CONTINUE',
            101: 'SWITCHING PROTOCOLS',
            200: 'OK',
            201: 'CREATED',
            202: 'ACCEPTED',
            203: 'NON-AUTHORITATIVE INFORMATION',
            204: 'NO CONTENT',
            205: 'RESET CONTENT',
            206: 'PARTIAL CONTENT',
            300: 'MULTIPLE CHOICES',
            301: 'MOVED PERMANENTLY',
            302: 'FOUND',
            303: 'SEE OTHER',
            304: 'NOT MODIFIED',
            305: 'USE PROXY',
            306: 'RESERVED',
            307: 'TEMPORARY REDIRECT',
            400: 'BAD REQUEST',
            401: 'UNAUTHORIZED',
            402: 'PAYMENT REQUIRED',
            403: 'FORBIDDEN',
            404: 'NOT FOUND',
            405: 'METHOD NOT ALLOWED',
            406: 'NOT ACCEPTABLE',
            407: 'PROXY AUTHENTICATION REQUIRED',
            408: 'REQUEST TIMEOUT',
            409: 'CONFLICT',
            410: 'GONE',
            411: 'LENGTH REQUIRED',
            412: 'PRECONDITION FAILED',
            413: 'REQUEST ENTITY TOO LARGE',
            414: 'REQUEST-URI TOO LONG',
            415: 'UNSUPPORTED MEDIA TYPE',
            416: 'REQUESTED RANGE NOT SATISFIABLE',
            417: 'EXPECTATION FAILED',
            500: 'INTERNAL SERVER ERROR',
            501: 'NOT IMPLEMENTED',
            502: 'BAD GATEWAY',
            503: 'SERVICE UNAVAILABLE',
            504: 'GATEWAY TIMEOUT',
            505: 'HTTP VERSION NOT SUPPORTED',
        }

    def __init__(self, stdout, stderr):
        self.stdout = stdout
        self.stderr = stderr

    def flush(self):
        """Flush method."""
        self.stdout.flush()
        self.stderr.flush()

if sys.platform == "win32":

    from ctypes import windll, Structure, c_short, c_ushort, byref

    SHORT = c_short
    WORD = c_ushort

    class Coord(Structure):
        """struct in wincon.h."""
        _fields_ = [
            ("X", SHORT),
            ("Y", SHORT)]

    class SmallRect(Structure):
        """struct in wincon.h."""
        _fields_ = [
            ("Left", SHORT),
            ("Top", SHORT),
            ("Right", SHORT),
            ("Bottom", SHORT)]

    class ConsoleScreenBufferInfo(Structure):
        """struct in wincon.h."""
        _fields_ = [
            ("dwSize", Coord),
            ("dwCursorPosition", Coord),
            ("wAttributes", WORD),
            ("srWindow", SmallRect),
            ("dwMaximumWindowSize", Coord)]

    # winbase.h
    STD_INPUT_HANDLE = -10
    STD_OUTPUT_HANDLE = -11
    STD_ERROR_HANDLE = -12

    # wincon.h
    FOREGROUND_BLACK     = 0x0000
    FOREGROUND_BLUE      = 0x0001
    FOREGROUND_GREEN     = 0x0002
    FOREGROUND_CYAN      = 0x0003
    FOREGROUND_RED       = 0x0004
    FOREGROUND_MAGENTA   = 0x0005
    FOREGROUND_YELLOW    = 0x0006
    FOREGROUND_GREY      = 0x0007
    FOREGROUND_INTENSITY = 0x0008 # foreground color is intensified.

    BACKGROUND_BLACK     = 0x0000
    BACKGROUND_BLUE      = 0x0010
    BACKGROUND_GREEN     = 0x0020
    BACKGROUND_CYAN      = 0x0030
    BACKGROUND_RED       = 0x0040
    BACKGROUND_MAGENTA   = 0x0050
    BACKGROUND_YELLOW    = 0x0060
    BACKGROUND_GREY      = 0x0070
    BACKGROUND_INTENSITY = 0x0080 # background color is intensified.

    STDOUT_HANDLE = windll.kernel32.GetStdHandle(STD_OUTPUT_HANDLE)
    SETCONSOLETEXTATTRIBUTE = windll.kernel32.SetConsoleTextAttribute
    GETCONSOLESCREENBUFFERINFO = windll.kernel32.GetConsoleScreenBufferInfo

    def get_text_attr():
        """Returns the character attributes (colors) of the console screen buffer."""
        csbi = ConsoleScreenBufferInfo()
        GETCONSOLESCREENBUFFERINFO(STDOUT_HANDLE, byref(csbi))
        return csbi.wAttributes

    def set_text_attr(color):
        """Sets the character attributes (colors) of the console screen buffer. Color is a combination of foreground
        and background color, foreground and background intensity."""
        SETCONSOLETEXTATTRIBUTE(STDOUT_HANDLE, color)

    class ColouredWriter(ColouredWriterBase):
        """Colour the Django server output."""
        colors = {  'endc': FOREGROUND_BLACK | FOREGROUND_INTENSITY,        # white
                    'fail': FOREGROUND_RED,                                 # red
                    'okgreen': FOREGROUND_GREEN,                            # green
                    'unknown': FOREGROUND_YELLOW,                           # yellow
                    'okblue': FOREGROUND_BLUE | FOREGROUND_INTENSITY,       # blue
                    'warning':FOREGROUND_MAGENTA | FOREGROUND_INTENSITY,    # magenta
                    'build': BACKGROUND_BLACK | BACKGROUND_INTENSITY | FOREGROUND_BLACK } # white

        #[14/Jul/2009 18:57:31] "GET /assets/maps/mp/q4ctf1.proc HTTP/1.1" 302 0
        line_re = re.compile('^(\[.*\]) (".*") (\d*) (\d*)$')
        build_re = re.compile("^(\[.*\]) ('.*')$")

        def __init__(self, stdout, stderr):
            ColouredWriterBase.__init__(self, stdout, stderr)
            self.default_colors = get_text_attr()
            self.default_bg = self.default_colors & 0x0070

        def write(self, line):
            """Write method."""
            line_m = ColouredWriter.line_re.match(line)
            if line_m:
                time = line_m.group(1)
                request = line_m.group(2)
                code = int(line_m.group(3))
                size = int(line_m.group(4))
                if code >= 500:
                    command = ColouredWriter.colors['fail']
                elif code >= 400:
                    command = ColouredWriter.colors['warning']
                elif code == 301:
                    command = ColouredWriter.colors['fail'] # We don't want any 301s from our links...
                elif code >= 300:
                    command = ColouredWriter.colors['okgreen']
                elif code >= 200:
                    command = ColouredWriter.colors['okblue']
                else:
                    command = ColouredWriter.colors['unknown']

                if code in ColouredWriter.status_code_text:
                    meaning = ColouredWriter.status_code_text[code]
                else:
                    meaning = "*unknown*"

                self.stdout.write(time)
                set_text_attr(command | self.default_bg)
                self.stdout.write(" " + request)
                set_text_attr(self.default_colors)
                self.stdout.write(" %i %i (%s)\n" % (code, size, meaning))
            else:
                build_m = ColouredWriter.build_re.match(line)
                if build_m:
                    time = build_m.group(1)
                    request = build_m.group(2)
                    self.stdout.write(time + ' ')
                    if request[1:].startswith("FAILED"):
                        set_text_attr(ColouredWriter.colors['fail'])
                    else:
                        set_text_attr(ColouredWriter.colors['build'] | self.default_bg)
                    self.stdout.write("%s" % (request))
                    set_text_attr(self.default_colors)
                    self.stdout.write("\n")
                else:
                    self.stdout.write(line)

else:

    class ColouredWriter(ColouredWriterBase):
        """Colour the Django server output."""
        colors = {  'endc': '\033[0m',               # black
                    'fail': '\033[91m',              # red
                    'okgreen': '\033[32m',           # green
                    'unknown': '\033[33m',           # yellow
                    'okblue': '\033[34m',            # blue
                    'warning': '\033[95m',           # magenta
                    'buildfail': '\033[91m\033[40m', # red on black
                    'buildmsg': '\033[32m\033[40m',  # green on black
                    'build': '\033[97m\033[40m' }    # white on black

        #[14/Jul/2009 18:57:31] "GET /assets/maps/mp/q4ctf1.proc HTTP/1.1" 302 0
        line_re = re.compile('^(\[.*\]) (".*") (\d*) (\d*)$')
        build_re = re.compile("^(\[.*\]) ('.*')$")

        # 127.0.0.1 0.0.0.0:8000 - [06/Sep/2009:21:40:00 +0100]
        #    "GET /assets/models/mapobjects/multiplayer/acceleration_pad/acceleration_pad_d.dds HTTP/1.1" 200 87125
        #    "http://0.0.0.0:8000/play/maps/mp/q4ctf1.map"
        #    "Mozilla/5.0 (Macintosh; U; Intel Mac OS X 10_6; en-us) AppleWebKit/532.0+ (KHTML, like Gecko) \
        #        Version/4.0.3 Safari/531.9"
        access_re = re.compile('^([\d\.]*) ([\w\.:-]*) - (\[.*\]) (".*") (\d*) (\d*) (".*") (".*")$')

        @classmethod
        def coloured_access(cls, time, request, code, size, server, who=None):
            """Generate a consistent coloured response."""
            if code >= 500:
                command = ColouredWriter.colors['fail']
            elif code >= 400:
                command = ColouredWriter.colors['warning']
            elif code == 301:
                command = ColouredWriter.colors['fail'] # We don't want any 301s from our links...
            elif code >= 300:
                command = ColouredWriter.colors['okgreen']
            elif code >= 200:
                command = ColouredWriter.colors['okblue']
            else:
                command = ColouredWriter.colors['unknown']

            if code in ColouredWriter.status_code_text:
                meaning = ColouredWriter.status_code_text[code]
            else:
                meaning = "*unknown*"

            endc = ColouredWriter.colors['endc']
            if who is None:
                line = "%s %s %s%s%s %i %i (%s)\n" % (server[:3], time, command, request, endc, code, size, meaning)
            else:
                line = "%s %s %s %s%s%s %i %i (%s)\n" % \
                    (server[:3], time, who, command, request, endc, code, size, meaning)
            return line

        def write(self, line):
            """Write method."""
            access_m = ColouredWriter.access_re.match(line)
            line_m = ColouredWriter.line_re.match(line)
            build_m = ColouredWriter.build_re.match(line)
            if access_m:
                server = "LIGHTTPD"
                time = access_m.group(3)
                who = access_m.group(1)
                request = access_m.group(4)
                code = int(access_m.group(5))
                size = int(access_m.group(6))
                line = ColouredWriter.coloured_access(time, request, code, size, server, who)
            elif line_m:
                server = "PYTHON"
                time = line_m.group(1)
                request = line_m.group(2)
                code = int(line_m.group(3))
                size = int(line_m.group(4))
                line = ColouredWriter.coloured_access(time, request, code, size, server)
            elif build_m:
                server = "BUILD"
                time = build_m.group(1)
                request = build_m.group(2)
                if request[1:].startswith("FAILED"):
                    build = ColouredWriter.colors['buildfail']
                elif request[1:].startswith("MSG"):
                    build = ColouredWriter.colors['buildmsg']
                else:
                    build = ColouredWriter.colors['build']
                endc = ColouredWriter.colors['endc']
                line = "%s %s %s%s%s\n" % (server[:3], time, build, request, endc)

            self.stdout.write(line)
            self.stdout.flush()

if __name__ == "__main__":
    CONVERTER = ColouredWriter(sys.stdout, sys.stderr)
    while True:
        L = sys.stdin.readline()
        CONVERTER.write(L)



# Copyright (c) 2012-2013 Turbulenz Limited

# Must be kept in sync with the version in the plugin and webgl engine
# code.  Python tools use this to stamp build products and to generate
# version checking code.
SDK_VERSION = "0.26.0.0"
